from aiddl_core.function.function import FunctionMixin, LazyFunctionMixin
from aiddl_core.representation.sym import Sym
from aiddl_core.representation.sym import Boolean
from aiddl_core.representation.int import Int
from aiddl_core.representation.tuple import Tuple
from aiddl_core.representation.collection import Collection
from aiddl_core.representation.set import Set
from aiddl_core.representation.substitution import Substitution


class SubstitutionFunction(FunctionMixin):
    def __call__(self, x):
        s = Substitution.from_term(x[1])
        return x[0].substitute(s)


class Quote(FunctionMixin, LazyFunctionMixin):
    def __call__(self, x):
        return x


class EvalRef(FunctionMixin):
    def __init__(self, evaluator):
        self.evaluator = evaluator

    def __call__(self, x):
        return self.evaluator(x)


class Matches(FunctionMixin):
    def __call__(self, args: Tuple) -> Boolean:
        x = args.get(0)
        y = args.get(1)
        sub = x.match(y)
        return Boolean.create(sub is not None)


class ExpandDomain(FunctionMixin, LazyFunctionMixin):
    MinKey = Sym("min")
    MaxKey = Sym("max")
    IncKey = Sym("inc")

    def __init__(self, evaluator):
        self.evaluator = evaluator

    def __call__(self, x):
        if not isinstance(x, Collection):
            return x
        return Set(self._eval_domain(x))

    def _eval_domain(self, C):
        domain = set()

        if C.contains_key(self.MinKey):
            min_value = self.evaluator(C.get(self.MinKey)).int_value
            max_value = self.evaluator(C.get(self.MaxKey)).int_value
            inc = self.evaluator(C.get_or_default(self.IncKey,
                                                  Int(1))).int_value
            for i in range(min_value, max_value+1, inc):
                domain.add(Int(i))
        else:
            for e in C:
                e = self.evaluator(e)
                if isinstance(e, Collection):
                    domain.addAll(self._eval_domain(e))
                else:
                    domain.add(e)
        return domain


class Let(FunctionMixin, LazyFunctionMixin):
    def __init__(self, evaluator):
        self.evaluator = evaluator

    def __call__(self, x):
        s = Substitution()
        for kvp in x.get(0):
            s.add(kvp.key, self.evaluator(kvp.value))
        return self.evaluator(x.get(1).substitute(s))


class Match(FunctionMixin, LazyFunctionMixin):
    def __init__(self, evaluator):
        self.evaluator = evaluator

    def __call__(self, x):
        if len(x) == 3:
            from_term = self.evaluator(x.get(0))
            to_term = self.evaluator(x.get(1))
            match_constraint = x.get(2)

            s = from_term.match(to_term)

            if s is None:
                return Boolean.create(False)

            return self.evaluator(match_constraint.substitute(s))
        else:
            p = x[0]
            for matchCase in x[1]:
                s = matchCase[0].match(p)
                if s is not None:
                    return self.evaluator(matchCase[1].substitute(s))
            raise ValueError("Match error:", x)


class EvalReferenceFunction(FunctionMixin):
    def __init__(self, evaluator):
        self.evaluator = evaluator

    def __call__(self, x):
        r = self.evaluator(x)
        return r


class EvalAllReferencesFunction(FunctionMixin):
    def __init__(self, evaluator):
        self.evaluator = evaluator

    def __call__(self, x):
        self.evaluator.set_follow_references(True)
        r = self.evaluator(x)
        self.evaluator.set_follow_references(False)
        return r


from aiddl_core.function.function import LazyFunctionMixin, FunctionMixin
from aiddl_core.representation.list import List
from aiddl_core.representation.set import Set
from aiddl_core.representation.sym import Sym
from aiddl_core.representation.tuple import Tuple
import aiddl_core.function as function


class Map(FunctionMixin, LazyFunctionMixin):
    def __init__(self, freg):
        self.freg = freg
        self.evaluator = freg.get_function(function.EVAL)

    def __call__(self, args):
        f = self.evaluator(args[0]).get_function_or_panic()
        C = self.evaluator(args[1])

        C_m = []
        for e in C:
            C_m.append(f(e))

        if isinstance(C, List):
            return List(C_m)
        else:
            return Set(C_m)


class Filter(FunctionMixin, LazyFunctionMixin):
    def __init__(self, freg):
        self.freg = freg
        self.evaluator = freg.get_function(function.EVAL)

    def __call__(self, args):
        f = self.evaluator(args[0]).get_function_or_panic()
        C = self.evaluator(args[1])
        C_m = []
        for e in C:
            if f(e).bool:
                C_m.append(e)

        if isinstance(C, List):
            return List(C_m)
        else:
            return Set(C_m)


class Reduce(FunctionMixin, LazyFunctionMixin):
    INIT_VAL = Sym("initial-value")

    def __init__(self, freg):
        self.freg = freg
        self.evaluator = freg.get_function(function.EVAL)

    def __call__(self, args):
        f = self.evaluator(args[0]).get_function_or_panic()
        C = self.evaluator(args[1])
        current_value = args.get(Reduce.INIT_VAL)

        for e in C:
            if current_value is None:
                current_value = e
            else:
                current_value = f(Tuple([current_value, e]))
        return current_value


class CallFunction(FunctionMixin):
    def __init__(self, freg):
        self.freg = freg

    def __call__(self, args):
        f = args[0]
        f_args = args[1]
        return f.get_function_or_panic()(f_args)


import aiddl_core.representation.num as numerical


class NaN(numerical.Num):
    @staticmethod
    def nan():
        return NaN()

    def __init__(self):
        pass

    def resolve(self, container):
        return self

    def __add__(self, other):
        return self

    def __sub__(self, other):
        return self

    def __mul__(self, other):
        return self

    def __truediv__(self, other):
        return self

    def __floordiv__(self, other):
        return self

    def __str__(self):
        return "NaN"

    def __eq__(self, other):
        return False

    def __ne__(self, other):
        return True

    def __lt__(self, other):
        return False

    def __le__(self, other):
        return False

    def __gt__(self, other):
        return False

    def __ge__(self, other):
        return False

    def is_zero(self):
        return False

    def is_positive(self):
        return False

    def is_negative(self):
        return False

    def is_nan(self):
        return True

    def __hash__(self):
        return 17*hash('NaN')

    def unpack(self):
        return float('NaN')



import aiddl_core.representation.term as term
from aiddl_core.representation.substitution import Substitution


class Var(term.Term):
    __slots__ = ["_name", "_is_anonymous"]
    __anon_count = 0

    def __init__(self, name=None):
        if name is None:
            Var.__anon_count += 1
            name = "?_X%d" % (Var.__anon_count)
            is_anon = True
        else:
            if name[0] != "?":
                name = "?" + name
            is_anon = False
        super(term.Term, self).__setattr__("_name", name)
        super(term.Term, self).__setattr__("_is_anonymous", is_anon)

    @property
    def is_anonymous(self):
        return self._is_anonymous

    def match(self, other):
        s = Substitution()
        s.add(self, other)
        return s

    def unpack(self):
        return str(self)

    def __str__(self):
        if self._is_anonymous:
            return "_"
        else:
            return str(self._name)

    def __repr__(self):
        return str(self)

    def __hash__(self):
        return 5*hash(self._name)

    def __eq__(self, other):
        if isinstance(other, Var):
            return self._name == other._name 
        else:
            return False

    def __ne__(self, other):
        return not self == other



import aiddl_core.representation.term as term
from aiddl_core.representation.collection import Collection
from aiddl_core.representation.keyval import KeyVal
from aiddl_core.representation.substitution import Substitution


class List(Collection):
    __slots__ = ["_internal_list", "_internal_map", "_hash"]

    def __init__(self, *args):
        internal_list = []
        internal_map = {}
        for t in args:
            if not isinstance(t, term.Term):
                for e in t:
                    internal_list.append(e)
                    if isinstance(e, KeyVal):
                        internal_map[e.key] = e.value
            else:
                internal_list.append(t)
                if isinstance(t, KeyVal):
                    internal_map[t.key] = t.value
        internal_list = tuple(internal_list)
        super(term.Term, self).__setattr__("_internal_list", internal_list)
        super(term.Term, self).__setattr__("_internal_map", internal_map)
        super(term.Term, self).__setattr__("_next_id", -1)
        super(term.Term, self).__setattr__("_hash", 3*hash(internal_list))

    def resolve(self, container):
        l_new = []
        change = False
        for e in self._internal_list:
            e_new = e.resolve(container)
            l_new.append(e_new)
            change = change or (e != e_new)
        if not change:
            return self
        else:
            return List(l_new)

    def substitute(self, substitution):
        l_new = []
        change = False
        for e in self._internal_list:
            e_new = e.substitute(substitution)
            l_new.append(e_new)
            change = change or (e != e_new)
        if not change:
            return self
        else:
            return List(l_new)

    def match(self, other):
        if isinstance(other, List):
            sCombined = Substitution()
            if len(self._internal_list) != len(other._internal_list):
                return None
            for i in range(len(self._internal_list)):
                subArg = self._internal_list[i].match(other._internal_list[i])
                if subArg is None:
                    return None
                if not sCombined.add_substitution(subArg):
                    return None
            return sCombined
        return None

    def __iter__(self):
        super(term.Term, self).__setattr__("_next_id", -1)
        return self

    def __next__(self):
        super(term.Term, self).__setattr__("_next_id", self._next_id + 1)
        if self._next_id == len(self._internal_list):
            super(term.Term, self).__setattr__("_next_id", -1)
            raise StopIteration
        else:
            return self._internal_list[self._next_id]

    def append_right(self, v):
        l_new = []
        for t in self._internal_list:
            l_new.append(t)
        l_new.append(v)
        return List(l_new)

    def append_left(self, v):
        l_new = []
        l_new.append(v)
        for t in self._internal_list:
            l_new.append(t)
        return List(l_new)

    def __add__(self, l):
        return self.add(l)
        
    def add(self, l):
        l_new = []
        for t in self._internal_list:
            l_new.append(t)
        l_new.append(l)
        return List(l_new)

    def add_all(self, l):
        l_new = []
        for t in self._internal_list:
            l_new.append(t)
        l_new += l
        return List(l_new)

    def put(self, key, value):
        l_new = []
        for t in self._internal_list:
            if not isinstance(t, KeyVal) or t.get_key() != key:
                l_new.add(t)
        l_new.add(KeyVal(key, value))
        return List(l_new)

    def put_all(self, values):
        l_new = []
        for k in self._internal_map.keys():
            if not values.contains_key(k):
                l_new.append(KeyVal(k, self.get(k)))
        for t in values:
            if isinstance(t, KeyVal):
                l_new.append(t)
        return List(l_new)

    def __len__(self):
        return len(self._internal_list)

    def size(self):
        return len(self._internal_list)

    def is_empty(self):
        return len(self._internal_list) == 0

    def remove(self, e):
        l_new = []
        for t in self._internal_list:
            if not e == t:
                l_new.append(t)
        return List(l_new)

    def remove_all(self, c):
        l_new = []
        for t in self._internal_list:
            if t not in c:
                l_new.append(t)
        return List(l_new)

    def get(self, n):
        if isinstance(n, int):
            return self._internal_list[n]
        if n in self._internal_map.keys():
            return self._internal_map[n]
        return None

    def is_unique_map(self):
        seen = set()
        for e in self._internal_list:
            if not isinstance(e, KeyVal) or e.get_key() in seen:
                return False
            seen.add(e.get_key())
        return True
    
    def __contains__(self, other):
        return other in self._internal_list

    def contains_key(self, other):
        return other in self._internal_map.keys()

    def __eq__(self, other):
        return isinstance(other, List) \
            and self._internal_list == other._internal_list

    def __ne__(self, other):
        return not self == other

    def __str__(self):
        if len(self._internal_list) == 0:
            return "[]"
        s = "["
        for e in self._internal_list:
            s += str(e) + " "
        return s[0:-1] + "]"

    def __repr__(self):
        return self.__str__()

    def __hash__(self):
        return self._hash

    def unpack(self):
        l = []
        for e in self:
            l.append(e.unpack())
        return l



from .sym import Sym
from .sym import Boolean
from .var import Var
from .collection import Collection
from .entref import EntRef
from .inf import Inf
from .int import Int
from .keyval import KeyVal
from .list import List
from .nan import NaN
from .num import Num
from .rat import Rat
from .real import Real
from .set import Set
from .str import Str
from .substitution import Substitution
from .term import Term
from .tuple import Tuple
from .funref import FunRef



import aiddl_core.representation.term as term
from aiddl_core.representation.sym import Sym
from aiddl_core.representation.tuple import Tuple
from aiddl_core.representation.entref import EntRef


class FunRef(term.Term):
    __slots__ = ["_fref", "_f", "_freg"]

    def __init__(self, fun_uri, fun_reg, module=None):
        if isinstance(fun_uri, EntRef):
            fun_name = fun_uri.target
            mod_name = fun_uri.module
            super(term.Term, self).__setattr__("_fref", mod_name + fun_name)
            super(term.Term, self).__setattr__("_freg", fun_reg)
        elif isinstance(fun_uri, Sym):
            if module is not None:
                super(term.Term, self).__setattr__("_fref", module + fun_uri)
            else:
                super(term.Term, self).__setattr__("_fref", fun_uri)
            super(term.Term, self).__setattr__("_freg", fun_reg)

        else:
            raise ValueError("%s is not Symbolic or Reference term" % str(fun_uri))

    def __call__(self, arg):
        """ Call the referenced function on a given argument

        :param arg: term arguments to the function
        :return: result of function application
        """
        return self._freg.get_function(self._fref)(arg)

    @property
    def function(self):
        """ Get the function referenced by this term

        :return: function
        """
        return self._freg.get_function(self._fref)

    def get_function_or_panic(self):
        """ Get function if it exists or raise an exception

        :return: function
        """
        return self._freg.get_function_or_panic(self._fref)

    @property
    def function_uri(self):
        return self._fref

    def substitute(self, s):
        return FunRef(self._fref.substitute(s), self._freg)

    def resolve(self, container):
        return self

    def __str__(self):
        return "^%s" % (str(self._fref))

    def __eq__(self, other):
        if isinstance(other, FunRef):
            return self._fref == other._fref
        return False

    def __ne__(self, other):
        return not (self == other)

    def __hash__(self):
        return 17*hash(self._fref)

    def unpack(self):
        return str(self)



from typing import Optional

from aiddl_core.representation.sym import Sym
from aiddl_core.representation.substitution import Substitution
from aiddl_core.representation.entref import EntRef
from aiddl_core.representation.tuple import Tuple
from aiddl_core.container.entry import Entry
from aiddl_core.container.module import Module
from aiddl_core.function import EVAL
import aiddl_core.function.default as dfun
from aiddl_core.util.logger import Logger

MOD = Sym("#mod")


class Container:
    def __init__(self):
        """ Create a new container

        """
        self.modules = {}
        self.module_list = []
        self.alias_lookup = {}
        self.self_alias_lookup = {}
        self.working_module = None
        self._fun_reg = dfun.get_default_function_registry(self)

    @property
    def fun_reg(self):
        """ Function registry used by this container

        :return: function registry
        """
        return self._fun_reg

    def eval(self, term):
        """ Evaluate a term

        :param term: term to evaluate
        :return: the term after all it has been fully evaluated
        """
        return self._fun_reg.get_function(EVAL)(term)

    @property
    def evaluator(self):
        """ Get the evaluator registered in the container function registry

        :return: the evaluator
        """
        return self._fun_reg.get_function(EVAL)

    def get_entry(self, name, module=None) -> Optional[Entry]:
        """ Get an entry stored in the container.

        If no module is specified, the working module will be used.

        :param name: name of the entry
        :param module: optional module
        :return: the entry if it exists, None otherwise
        """
        if module is None:
            m = self.modules[self.working_module]
        else:
            if module in self.modules.keys():
                m = self.modules[module]
            else:
                m = None
        if m is None:
            print("Registered Modules:")
            for m_e in self.module_list:
                print(m_e.name)
                print(type(m_e.name))
                print(self.modules[m_e.name] is not None)
                print(m_e.name == module)
            raise AttributeError(
                f"Requesting entry from unknown module{module} with type {type(module)}")
        for e in m.get_entries():
            s = e.name.match(name)
            if s is not None:
                return e.substitute(s)
        return None

    def get_processed_value(self, name, module=None) -> Optional['Term']:
        """ Get entry found under name and return its fully processed value

        If no module is specified, the working module will be used.

        :param name: name of the entry
        :param module: optional module
        :return: processed value or None if entry does not exist
        """
        e = self.get_entry(name, module=module)
        if e is None:
            return e
        else:
            self.eval.set_follow_references(True)
            r = self.eval(e.get_value().resolve(self))
            self.eval.set_follow_references(False)
            return r

    def get_processed_value_or_panic(self, name, module=None):
        """ Get entry found under name and return its fully processed value. Raises an exception if entry does not
        exist.

        If no module is specified, the working module will be used.

        :param name: name of the entry
        :param module: optional module
        :return: processed value
        """
        self.evaluator.set_follow_references(True)
        r = self.eval(self.get_entry(name, module=module).value.resolve(self))
        self.evaluator.set_follow_references(False)
        return r

    def get_matching_entries(self, module_pattern, type_pattern, name_pattern):
        """ Collect all entries with matching modules, types, and names.

        :param module_pattern: term to match to modules
        :param type_pattern: term to match to types
        :param name_pattern: term to match to names
        :return: list of all entries with matching module, type, and name
        """
        r = []
        for m in self.module_list:
            if module_pattern is not None:
                s_base = module_pattern.match(m.name)
            else:
                s_base = Substitution()
            if s_base is not None:
                for e in m.get_entries():
                    s = s_base.copy()

                    if type_pattern is not None:
                        s_tmp = type_pattern.match(e.type)
                        if s_tmp is None or not s.add_substitution(s_tmp):
                            continue
                    if name_pattern is not None:
                        s_tmp = name_pattern.match(e.name)
                        if s_tmp is None or not s.add_substitution(s_tmp):
                            continue
                        else:
                            r.append(e)
                    elif s is not None:
                        r.append(e)
        return r

    def set_entry(self, entry, module=None):
        """ Set an entry in a module.

         If no module is specified, the working module will be used.

        :param entry: the entry
        :param module: optional module
        """
        if module is not None:
            m_set = self.modules[module]
        else:
            m_set = self.modules[self.working_module]

        if m_set is None:
            raise ValueError(f"Module {module} does not exist.")
        if entry.name.resolve(self) != entry.name:
            raise ValueError(f"Entry name {entry.name}"
                             + " has references. Make sure to resolve"
                             + " references before creating the entry.")
        prev = self.get_entry(entry.name, module=module)
        if prev is not None:
            m_set.remove_entry(prev)
        m_set.put_entry(entry)

    def copy_entry(self, a, module_a, b, module_b):
        """ Copy an entry (to another module)

        :param a: name of entry to copy
        :param module_a:  module of entry to copy
        :param b: name of entry copy
        :param module_b: module of entry copy
        """
        a_entry = self.get_entry(a, module=module_a)
        self.set_entry(Entry(a_entry.type,
                             b,
                             a_entry.value),
                       module=module_b)

    def delete_entry(self, entry, module=None):
        """ Delete an entry

        If no module is specified, the working module will be used.

        :param entry: entry to be removed
        :param module: optional module
        """
        if module is not None:
            m_set = self.modules[module]
        else:
            m_set = self.modules[self.working_module]
        if m_set is None:
            raise ValueError(f"Module {module} does not exist.")
        m_set.remove_entry(entry)

    def export(self, module_name, filename):
        """ Export a module to a file.

        :param module_name: name of the module
        :param filename: filename to write to
        """
        m = self.modules.get(module_name)
        if m is None:
            raise ValueError(f"Requesting non-existing module: {module_name}")
        f_out = open(filename, "w")
#        f_out.write("(#mod self " + m.get_name())
#        f_out.write("\n")
        for e in m.get_entries():
            f_out.write("(")
            f_out.write(str(e.type))
            f_out.write("\n\t")
            f_out.write(Logger.pretty_print(e.name, 1))
            f_out.write("\n\t")
            f_out.write(Logger.pretty_print(e.value, 1))
            f_out.write("\n)\n\n")
        f_out.close()

    def resolve_reference(self, t):
        """ Resolve an entry reference by replacing it with its referent until it becomes a non entry reference term.

        :param t:
        :return:
        :raises: ValueError if reference cannot be resolved or resolving references leads to loop
        """
        next_term = t
        seen = []
        while True:
            seen.append(next_term)
            if isinstance(next_term, EntRef):
                next_term = self._resolve_reference_once(next_term)
                if next_term in seen:
                    raise ValueError(f"Infinite reference loop {seen} cannot be fully resolved.")
            else:
                return next_term
            if next_term is None:
                break
        raise ValueError("Reference cannot be resolved:", t)

    def _resolve_reference_once(self, t):
        if isinstance(t.target, EntRef):
            return t.target
        e = self.get_entry(t.target, module=t.module)
        if e is None:
            return None
        return e.value

    def get_working_module(self) -> Sym:
        """ Get the name of the working module

        :return: name of working module
        """
        return self.working_module

    def set_working_module(self, module: Sym):
        """ Set the working module

        :param module: name of working module
        """
        if module not in self.modules.keys():
            raise ValueError(f"Trying use non-existing module {module} as working module.")
        self.working_module = self.modules[module].name

    def add_module(self, name: Sym):
        """ Add a new module.

        The first added module becomes the working module

        :param name: name of module
        """
        if not isinstance(name, Sym):
            raise ValueError(f"Illegal module name: {name}. Name must be of type Sym.")
        m = Module(name)
        if name not in self.modules.keys():
            self.modules[name] = m
            self.module_list.append(m)
        if self.working_module is None:
            self.working_module = m.name

    def remove_module(self, name: Sym):
        """ Remove a module

        :param name: name of module
        """
        m = self.modules[name]
        del self.modules[name]
        self.module_list.remove(m)
        if self.working_module == m:
            self.working_module = None

    def get_module_names(self):
        """ Get list of all module names

        :return: list of module names
        """
        r = []
        for m in self.modules.keys():
            r.append(m)
        return r

    def add_module_alias(self, source_module, alias, target_module):
        """ Add an alias for a module used by another module.

        :param source_module: module that uses the alias
        :param alias: the alias
        :param target_module: module that the alias refers to
        """
        if source_module not in self.alias_lookup.keys():
            self.alias_lookup[source_module] = {}
        self.alias_lookup[source_module][alias] = target_module
        if source_module == target_module:
            self.self_alias_lookup[source_module] = alias

    def resolve_module_alias(self, source_module, alias) -> Optional[Sym]:
        """ Resolve a module alias

        :param source_module: module that uses the alias
        :param alias: the alias
        :return: URI of target module or None if alias does not exist
        """
        if source_module in self.alias_lookup.keys():
            return self.alias_lookup[source_module][alias]
        return None

    def find_self_alias(self, module_name):
        """ Get the name a module uses to refer to itself.

        Specified with the name of the #mod entry in every module.
        For example (#mod self module-uri) has the self alias 'self'.


        :param module_name: name of the module
        :return: self alias
        """
        return self.self_alias_lookup.get(module_name)

    # def register_observer(self, module, entryName, obs):
    #     m = self.modules.get(module)
    #     if m is None:
    #         raise ValueError(
    #             "Trying to add observer to non-existing module: " + module)
    #     m.addObserver(entryName, obs)

    def toggle_namespaces(self, flag):
        """ Toggle namespaces on or off.

        This will lead to substitutions in all entries of modules that use namespaces

        :param flag: true to turn namespace on, false to rever changes made by namespaces
        """
        sub_map = {}
        for m in self.module_list:
            ns_sub = Substitution()
            for e in self.modules[m.name].get_entries():
                if not e.type == MOD:
                    ns_sub.add(e.name, e.value)
                    if not flag:
                        ns_sub = ns_sub.inverse()
                    sub_map[m.name] = ns_sub

        for m in self.module_list:
            namespaces = self.get_matching_entries(m.name,
                                                   Sym("#namespace"),
                                                   None)
            namespaces += self.get_matching_entries(m.name,
                                                    Sym("#nms"),
                                                    None)
            s = Substitution()
            for ns in namespaces:
                ns_term = ns.value

                if isinstance(ns_term, Sym):
                    print(f"[Warning] Deprecated #nms/#namespace usage: {ns}."
                          f" Use (reference to) set of key-values instead.")
                    ns_mod_name = self.resolve_module_alias(m.name,
                                                            ns.name)

                    if ns_mod_name is None:
                        raise ValueError("Could not find alias: %s %s" %
                                         (m.name, ns.name))
                    if not s.add_substitution(sub_map.get(ns_mod_name)):
                        raise ValueError("Namespace conflict in module "
                                         + str(m.name)
                                         + ". Change namespace or use "
                                         + "#req and references.")

                elif isinstance(ns_term, Tuple):
                    print(f"[Warning] Deprecated #nms/#namespace usage: {ns}."
                          f" Use (reference to) set of key-values instead.")
                    ns_name = ns_term[0]
                    ns_table = self.resolve_reference(ns_term[1])
                    ns_sub = Substitution()
                    for entry in ns_table:
                        ns_sub.add(entry[ns_name], entry[0])
                    if not s.add_substitution(ns_sub):
                        print("Namespace term:", ns_term)
                        print("Substitution before:", s)
                        print("Failed to add:", ns_sub)
                        raise ValueError(f"Namespace conflict in module {m.name}. "
                                         f"Change namespace or use #req and references.")
                else:
                    ns_res = ns_term.resolve(self)
                    ns_sub = Substitution.from_term(ns_res)
                    if not s.add_substitution(ns_sub):
                        print("Namespace term:", ns_term)
                        print("Substitution before:", s)
                        print("Failed to add:", ns_sub)
                        raise ValueError(f"Namespace {ns.name} creates conflict in module {m.name}."
                                         f"Change namespace or use #req and references.")

            if len(s) > 0:
                m.substitute(s)

    def __str__(self):
        s = ""
        s += "Entries:\n"
        for m in self.module_list:
            s += "Module: " + str(m._module_uri)
            if m == self.working_module:
                s += " <--- Working Module"
            s += "\n"
            for e in m.get_entries():
                s += "\t"
                s += str(e)
                s += "\n"
        return s



import aiddl_core.representation
from aiddl_core.representation import term


class Sym(term.Term):
    __slots__ = ["_value"]

    def __init__(self, value):
        # if value == "+INF":
        #     raise ValueError("This should be infinity")
        if not isinstance(value, str):
            raise ValueError("Symbolic must be instanciated with a string.")
        super(term.Term, self).__setattr__("_value", value)

    @property
    def string(self) -> str:
        """ Get the string of this value

        :return: string value of this symbol
        """
        return self._value

    @property
    def bool(self) -> bool:
        """ Get the boolean value of this term if it is true or false

        :return: the boolean value of this term
        """
        if self._value == "true":
            return True
        elif self._value == "false":
            return False
        raise AttributeError("Not a boolean value", self)

    def __add__(self, other):
        if isinstance(other, Sym):
            return Sym(self._value + "." + other._value)
        raise AttributeError("Not a Symbolic: " + str(other)
                             + " (concat only defined between symbolic terms")

    def __str__(self):
        return str(self._value)

    def __repr__(self):
        return str(self._value)

    def __hash__(self):
        return 3*hash(self._value)

    def __eq__(self, other):
        if isinstance(other, Sym):
            return self._value == other._value
        return False

    def __ne__(self, other):
        return not self == other

    def unpack(self):
        return str(self)


class Boolean(Sym):
    __slots__ = ["_bool_value", "_value"]

    def __init__(self, value):
        if value:
            super(term.Term, self).__setattr__("_value", "true")
            super(term.Term, self).__setattr__("_bool_value", True)
        else:
            super(term.Term, self).__setattr__("_value", "false")
            super(term.Term, self).__setattr__("_bool_value", False)

    @property
    def bool(self):
        """ Get the boolean value of this term

        :return: the boolean value
        """
        return self._bool_value

    @staticmethod
    def create(val: bool):
        """ Create a boolean term from a bool object

        :param val: boolean value
        :return: boolean term
        """
        if val:
            return TRUE
        else:
            return FALSE

    def __or__(self, other):
        """ Logical or between this Boolean term and another Boolean term

        :param other: Boolean term
        :return: self or other
        """
        return Boolean.create(self.bool or other.bool)

    def __and__(self, other):
        """ Logical and between this Boolean term and another Boolean term

        :param other: Boolean term
        :return: self and other
        """
        return Boolean.create(self.bool and other.bool)

    def __str__(self):
        return str(self._value)

    def __repr__(self):
        return str(self._value)

    def unpack(self):
        return self._bool_value


TRUE = Boolean(True)
FALSE = Boolean(False)



from abc import ABC, abstractmethod
from typing import Optional, Union

# from aiddl_core.container.container import Container
from aiddl_core.representation.substitution import Substitution


class Term(ABC):
    def size(self) -> int:
        """ Get size if this is a collection term or tuple

        :return: length of term if applicable
        :raises: AttributeError if length not defined for this term
        """
        return len(self)

    def __len__(self) -> int:
        """ Get size if this is a collection term or tuple

        :return: length of term if applicable
        :raises: AttributeError if length not defined for this term
        """
        raise AttributeError("Not implemented for: " +
                             str(type(self)) + " " + str(self))

    def __call__(self, x: Union[int, 'Term']) -> Optional['Term']:
        """ Apply x to this term (see get(x) below)

        :param x: integer index or term
        :return: term resulting from applying x to this term if defined, None if the
                 operation is supported but the result undefined
        :raises: attribute error if this term does not support the operation
        """
        self.get(x)

    def __getitem__(self, x: Union[int, 'Term']) -> Optional['Term']:
        """ Apply x to this term (see get(x) below)

        :param x: integer index or term
        :return: term resulting from applying x to this term if defined, None if the
                 operation is supported but the result undefined
        :raises: attribute error if this term does not support the operation
        """
        return self.get(x)

    def get(self, x: Union[int, 'Term']) -> Optional['Term']:
        """ Apply x to this term and get the result

        If this term is a collection or tuple and x is a term, this method will look for a
        key-value pair with key x. If this term is a tuple or list and x is an integer,
        this method will look for the value at index x. If this term is a function reference
        to a function f, this method will return f(x). Otherwise, an exception is raised.

        :param x: integer index or term
        :return: term resulting from applying x to this term if defined, None if the
                 operation is supported but the result undefined
        :raises: attribute error if this term does not support the operation
        """
        raise AttributeError(f"Not implemented for: {type(self)} {self}")

    def get_or_default(self, x, default) -> 'Term':
        """ Use self.get(x) and return a default value if the result is None

        :param x: integer index or term
        :param default: a default value
        :return: result term or default
        """
        result = self.get(x)
        if result is not None:
            return result
        else:
            return default

    def get_or_panic(self, x, default):
        """ Use self.get(x) and throw an exception if the result is None

        :param x:
        :param default:
        :return:
        """
        result = self.get(x)
        if result is not None:
            return result
        raise AttributeError(f"Key {type(self)} not found in {self}")

    @abstractmethod
    def unpack(self) -> object:
        """  Unpack AIDDL term to regular python object.

        :return: a python object version of this AIDDL term
        """

    def match(self, other: 'Term') -> Optional['Substitution']:
        """ Attempt to match this term to another

        :param other: another term
        :return: a substitution that makes this term equal to the other one or None if this is not possible
        """
        if self == other:
            return Substitution()
        return None

    def resolve(self, container: 'Container'):
        """ Recursively resolve any references found in this term

        :param container: Used to look up the values of any referenced entries
        :return: Term with all references resolved
        """
        return self

    def substitute(self, substitution: 'Substitution') -> 'Term':
        """ Apply a substitution to this term and return the result.

        :param substitution: collection of replacement terms
        :return: term with all replacements applied
        """
        return substitution.substitute(self)

    @abstractmethod
    def __eq__(self, other):
        """Check if two terms are equal."""

    @abstractmethod
    def __ne__(self, other):
        """Check if two terms are not equal."""

    @abstractmethod
    def __str__(self, other):
        """Create string from term."""

    def __setattr__(self, name, value):
        raise AttributeError("Immutable object.")

    def __setitem__(self, key, value):
        raise AttributeError("Immutable object.")



import aiddl_core.representation.term as term
from aiddl_core.representation import Sym


class EntRef(term.Term):
    """An entry reference used to refer to terms in other entries."""
    __slots__ = ["_ref_target", "_alias", "_mod_name"]

    def __init__(self, ref_target, mod_name, alias=None):
        super(term.Term, self).__setattr__("_ref_target", ref_target)
        super(term.Term, self).__setattr__("_mod_name", mod_name)
        super(term.Term, self).__setattr__("_alias", alias)

    @property
    def target(self) -> term.Term:
        """ Target of the reference (i.e. the name of an entry)

        :return: name of target entry
        """
        return self._ref_target

    @property
    def module(self) -> Sym:
        """ Module of the reference entry

        :return: name of the module
        """
        return self._mod_name

    @property
    def alias(self) -> Sym:
        """ Alias of the reference entry

        :return: alias of the module
        """
        return self._alias

    def resolve(self, container):
        return container.resolve_reference(self)

    def convert2uri(self):
        """ Attempt to convert this entry reference to a symbolic URI

        Concatenates the module name with the target. Only works if the target is symbolic.

        :return: New symbol made of the module name and the target
        """
        return self._mod_name + self._ref_target

    def substitute(self, s):
        if self._alias is None:
            return EntRef(self._ref_target.substitute(s),
                          self._mod_name.substitute(s))
        else:
            return EntRef(self._ref_target.substitute(s),
                          self._mod_name.substitute(s),
                          alias=self._alias.substitute(s))

    def __str__(self):
        if self._alias is not None:
            return "%s@%s" % (str(self._ref_target), str(self._alias))
        else:
            return "%s@%s" % (str(self._ref_target), str(self._mod_name))

    def __repr__(self):
        return self.__str__()
        
    def __eq__(self, other):
        if isinstance(other, EntRef):
            return self._ref_target == other._ref_target \
                and self._mod_name == other._mod_name \
                and self._alias == other._alias
        else:
            return False

    def __ne__(self, other):
        return not (self == other)

    def __hash__(self):
        return 3*hash(self._ref_target) \
            + 5*hash(self._alias) \
            + 7*hash(self._mod_name)

    def unpack(self):
        return str(self)
        



from aiddl_core.function.function import LazyFunctionMixin, FunctionMixin
from aiddl_core.representation.sym import Boolean
from aiddl_core.representation.collection import Collection
from aiddl_core.representation.tuple import Tuple
from aiddl_core.representation.list import List


class And(FunctionMixin, LazyFunctionMixin):
    def __init__(self, evaluator):
        self.evaluator = evaluator

    def __call__(self, x):
        if x.size() == 0:
            return Boolean.create(True)
        for i in range(1, x.size()):
            if not self.evaluator(x.get(i)).bool:
                return Boolean.create(False)
        return Boolean.create(True)


class If(FunctionMixin, LazyFunctionMixin):
    def __init__(self, evaluator):
        self.evaluator = evaluator

    def __call__(self, x):
        if self.evaluator(x.get(0)).bool:
            return self.evaluator(x.get(1))
        else:
            return self.evaluator(x.get(2))


class Or(FunctionMixin, LazyFunctionMixin):
    def __init__(self, evaluator):
        self.evaluator = evaluator

    def __call__(self, x):
        if x.size() == 0:
            return Boolean.create(False)
        for i in range(1, x.size()):
            if self.evaluator(x.get(i)).bool:
                return Boolean.create(True)
        return Boolean.create(False)


class Cond(FunctionMixin, LazyFunctionMixin):
    def __init__(self, evaluator):
        self.evaluator = evaluator

    def __call__(self, x):
        for i in range(0, x.size()):
            if self.evaluator(x.get(i).key).bool:
                return self.evaluator(x.get(i).value)
        raise ValueError("Conditional does not cover all cases.\n"
                         + str(x) + "\n"
                         + "Make sure the existing conditions cover all cases,"
                         + " or add a catch all case (True) : <default-result>"
                         + " at the end to avoid this.")


class Exists(FunctionMixin, LazyFunctionMixin):
    ExistsHelp = "Uses format: (exists (x S C)) where x is a term" \
                 + " matched to all elements of collection term s.\n" \
                 + "The resulting terms must satisfy all constraints in set C."

    def __init__(self, evaluator):
        self.evaluator = evaluator

    def __call__(self, x):
        if len(x) != 3:
            raise ValueError(x, ": ", Exists.ExistsHelp)

        matchTerm = x.get(0)
        collTerm = self.evaluator(x.get(1))

        if not isinstance(collTerm, Collection) and \
           not isinstance(collTerm, Tuple):
            raise ValueError(
                x.get(2) + " not a CollectionTerm or TupleTerm "
                + Exists.ExistsHelp)

        if isinstance(collTerm, Collection):
            collection = collTerm
        else:
            list = []
            for i in range(collTerm.size()):
                list.add(collTerm.get(i))
            collection = List(list)

        constraints = x.get(2)

        for t in collection:
            s = matchTerm.match(t)

            if s is None:
                continue

            conSub = constraints.substitute(s)

            if self.evaluator(conSub).bool:
                return Boolean.create(True)
        return Boolean.create(False)


class Forall(FunctionMixin, LazyFunctionMixin):
    ForallHelp = "Uses format: (forall (x S C)) where x is a term" \
                 + " matched to all elements of collection term S.\n" \
                 + "The resulting terms is then evaluated based on C."

    def __init__(self, evaluator):
        self.evaluator = evaluator

    def __call__(self, x):
        if len(x) != 3:
            raise ValueError(x, ":", Forall.ForallHelp)

        matchTerm = x.get(0)
        collTerm = self.evaluator(x.get(1))

        if not isinstance(collTerm, Collection) and \
           not isinstance(collTerm, Tuple):
            raise ValueError(
                x.get(2),
                " not a CollectionTerm or TupleTerm ",
                Forall.ForallHelp)

        if isinstance(collTerm, Collection):
            collection = collTerm
        else:
            list = []
            for i in range(collTerm.size()):
                list.add(collTerm.get(i))
            collection = List(list)

        constraints = x.get(2)

        for t in collection:
            s = matchTerm.match(t)
            if s is None:
                return Boolean.create(False)

            conSub = constraints.substitute(s)

            if not self.evaluator(conSub).bool:
                return Boolean.create(False)
        return Boolean.create(True)



from abc import ABC, abstractmethod

import aiddl_core.function as function
from aiddl_core.representation.funref import FunRef
from aiddl_core.representation.substitution import Substitution
from aiddl_core.representation.sym import Sym

SELF = Sym("#self")
SELF_ALT = Sym("#arg")


class FunctionMixin:
    @abstractmethod
    def __call__(self, args):
        """Apply function to input term."""


class LazyFunctionMixin:
    pass


class InitializableMixin:
    @abstractmethod
    def initialize(self, args):
        """Initialize function."""


class ConfigurableMixin:
    @abstractmethod
    def configure(self, cfg, freg):
        """Configure function."""


class InterfaceImplementationMixin:
    @abstractmethod
    def get_interface_uri(self):
        """Get function interface URI."""


class NamedFunction(FunctionMixin):
    def __init__(self, name, f, e, args=None):
        self.name = name
        self.f = f
        self.e = e
        self.args = args

    def __call__(self, arg):
        if self.args is None:
            s = Substitution()
            s.add(SELF, arg)
            s.add(SELF_ALT, arg)
        else:
            s = self.args.match(arg)
        return self.e(self.f.substitute(s))


class LambdaEvaluator(FunctionMixin, LazyFunctionMixin):
    NextID = 0

    def __init__(self, fun_reg):
        self._fun_reg = fun_reg
        self._evaluator = fun_reg.get_function(function.EVAL)

    def __call__(self, arg):
        arg_term = arg[0]
        fun_term = arg[1]

        f = LambdaFunction(arg_term, fun_term, self._evaluator)
        uri = Sym("#lambda_%d" % LambdaEvaluator.NextID)
        LambdaEvaluator.NextID += 1
        self._fun_reg.add_function(uri, f)
        return FunRef(uri, self._fun_reg)


class LambdaFunction(FunctionMixin):
    def __init__(self, x, f, e):
        self.x = x
        self.f = f
        self.e = e

    def __call__(self, arg):
        s = self.x.match(arg)
        return self.e(self.f.substitute(s))


class InitFunction(FunctionMixin):
    def __init__(self, fun_reg):
        self._fun_reg = fun_reg

    def __call__(self, x):
        f = x[0]
        args = x[1]

        if isinstance(f, Sym):
            func = self._fun_reg.get_function_or_panic(f)
            uri = f
        elif isinstance(f, FunRef):
            func = f.function
            uri = f.function_uri
        else:
            raise ValueError(f"Not a symbol or function reference: {f}")

        func.initialize(args)
        return FunRef(uri, self._fun_reg)


class ConfigFunction(FunctionMixin):
    def __init__(self, fun_reg):
        self._fun_reg = fun_reg

    def __call__(self, x):
        f = x[0]
        args = x[1]

        if isinstance(f, Sym):
            func = self._fun_reg.get_function_or_panic(f)
            uri = f
        elif isinstance(f, FunRef):
            func = f.function
            uri = f.function_uri
        else:
            raise ValueError(f"Not a symbol or function reference: {f}")

        func.configure(args, self._fun_reg)
        return FunRef(uri, self._fun_reg)



from aiddl_core.function.function import FunctionMixin, LazyFunctionMixin
from aiddl_core.representation.sym import Boolean
from aiddl_core.representation.int import Int
from aiddl_core.representation.num import Num
from aiddl_core.representation.tuple import Tuple
from aiddl_core.representation.collection import Collection
from aiddl_core.representation.list import List
from aiddl_core.representation.set import Set


class ContainsMatch(FunctionMixin):
    def __call__(self, args: Tuple) -> Boolean:
        c = args[0]
        e = args[1]
        for e_c in c:
            if e.match(e_c) is not None:
                return Boolean(True)
        return Boolean(False)


class ContainsAny(FunctionMixin):
    def __call__(self, args: Tuple) -> Boolean:
        c1 = args[0]
        c2 = args[1]
        for e in c2:
            if e in c1:
                return Boolean(True)
        return Boolean(False)


class Sum(FunctionMixin):
    def __call__(self, c: Collection) -> Num:
        s = Int(0)
        for e in c:
            s += e
        return s


class Union(FunctionMixin):
    def __call__(self, x: Collection) -> Set:
        u = set()
        for s in x:
            for e in s:
                u.add(e)
        return Set(u)


class Concat(FunctionMixin):
    def __call__(self, x):
        cl = []
        for s in x:
            for e in s:
                cl.append(e)
        return List(cl)


class Zip(FunctionMixin, LazyFunctionMixin):
    def __init__(self, evaluator):
        self.evaluator = evaluator

    def __call__(self, x):
        zip_term = self.evaluator(x)

        zip_list = []
        for t in zip_term:
            if isinstance(t, Tuple) or isinstance(t, List):
                zip_list.append(t)
            else:
                raise ValueError("Bad argument for zip: %s. Can only zip tuples or lists." % (str(t)))

        zipped_list = []
        idx = 0
        done = False
        while not done:
            zip_line = []
            for i in range(0, len(zip_list)):
                if idx >= len(zip_list[i]):
                    done = True
                    break
                zip_line.append(zip_list[i][idx])
            if not done:
                zipped_list.append(Tuple(zip_line))
                idx += 1
        return List(zipped_list)


import importlib

from aiddl_core.representation.funref import FunRef
from aiddl_core.representation.sym import Sym


class PythonFunctionLoader:
    def __init__(self, freg):
        self.freg = freg

    def __call__(self, x):
        name = x[Sym("name")]
        module = x[Sym("module")]
        py_class = x[Sym("pyclass")]
        py_module = x[Sym("pymodule")]

        imp = importlib.import_module(str(py_module))
        instance = getattr(imp, str(py_class))()
        uri = module + name

        self.freg.add_function(uri, instance)


class PythonFunctionFactoryLoader:
    def __init__(self, freg):
        self.freg = freg

    def __call__(self, x):
        name = x[Sym("name")]
        module = x[Sym("module")]
        py_class = x[Sym("pyclass")]
        py_module = x[Sym("pymodule")]

        imp = importlib.import_module(str(py_module))
        constructor = getattr(imp, str(py_class))

        factory = PythonFunctionFactory(constructor, self.freg)
        uri = module + name

        self.freg.add_function(uri, factory)
        return FunRef(uri, self.freg)


class PythonFunctionFactory:
    def __init__(self, constructor, freg):
        self.constructor = constructor
        self.freg = freg

    def __call__(self, x):
        uri = None
        init = None
        config = None
        if isinstance(x, Sym):
            uri = x
        else:
            uri = x[0]
            init = x[Sym("init")]
            config = x[Sym("config")]

        f = self.constructor()

        if init is not None:
            f.initialize(init)

        if config is not None:
            f.configure(config, self.freg)

        self.freg.add_function(uri, f)
        return FunRef(uri, self.freg)


import aiddl_core.representation.term as term
import aiddl_core.representation.num as numerical
import aiddl_core.representation.int as integer
import aiddl_core.representation.nan as nan


class Inf(numerical.Num):
    __slots__ = ["_is_positive"]

    @staticmethod
    def pos():
        """ Positive infinity

        :return: a +INF term
        """
        return Inf(True)

    @staticmethod
    def neg():
        """ Negative infinity

        :return: a -INF term
        """
        return Inf(False)

    def __init__(self, is_positive):
        """ Create a new infinite term

        :param is_positive: true for positive infinity, false for negative infinity
        """
        super(term.Term, self).__setattr__("_is_positive", is_positive)

    def __add__(self, other):
        if isinstance(other, Inf):
            if self._is_positive == other._is_positive:
                return self
            return nan.NaN()
        elif isinstance(other, numerical.Num):
            return self
        elif isinstance(other, nan.NaN):
            return other
        raise AttributeError("Not a numerical term: " + str(other))

    def __sub__(self, other):
        if isinstance(other, Inf):
            if self._is_positive != other._is_positive:
                return self
            return nan.NaN()
        elif isinstance(other, numerical.Num):
            return self
        elif isinstance(other, nan.NaN):
            return other
        raise AttributeError("Not a numerical term: " + str(other))

    def __mul__(self, other):
        if isinstance(other, Inf):
            if self._is_positive == other._is_positive:
                return Inf.pos()
            else:
                return Inf.neg()
        elif isinstance(other, numerical.Num):
            if other == integer.Int(0):
                return nan.NaN()
            else:
                if self._is_positive == (other > integer.Int(0)):
                    return Inf.pos()
                else:
                    return Inf.neg()
        elif isinstance(other, nan.NaN):
            return other
        raise AttributeError("Not a numerical term: " + str(other))

    def __truediv__(self, other):
        if isinstance(other, Inf):
            return nan.NaN()
        elif isinstance(other, numerical.Num):
            if self._is_positive == (other > integer.Int(0)):
                return Inf.pos()
            else:
                return Inf.neg()
        elif isinstance(other, nan.NaN):
            return other
        raise AttributeError("Not a numerical term: " + str(other))

    def __floordiv__(self, other):
        if isinstance(other, Inf):
            return nan.NaN()
        elif isinstance(other, numerical.Num):
            if self._is_positive == (other > integer.Int(0)):
                return Inf.pos()
            else:
                return Inf.neg()
        elif isinstance(other, nan.NaN):
            return other
        raise AttributeError("Not a numerical term: " + str(other))

    def __str__(self):
        if self._is_positive:
            return "+INF"
        else:
            return "-INF"

    def __eq__(self, other):
        return False

    def __ne__(self, other):
        return True

    def __lt__(self, other):
        if isinstance(other, Inf):
            if not self._is_positive and other.is_positive():
                return True
            else:
                return False
        if isinstance(other, numerical.Num):
            return not self._is_positive
        elif isinstance(other, nan.Nan):
            return False
        raise AttributeError("Not a numerical term: " + str(other))

    def __le__(self, other):
        if isinstance(other, Inf):
            if not self._is_positive:
                return True
            else:
                return False
        if isinstance(other, numerical.Num):
            return not self._is_positive
        elif isinstance(other, nan.Nan):
            return False
        raise AttributeError("Not a numerical term: " + str(other))

    def __gt__(self, other):
        if isinstance(other, Inf):
            if self._is_positive and not other.is_positive():
                return True
            else:
                return False
        if isinstance(other, numerical.Num):
            return self._is_positive
        elif isinstance(other, nan.Nan):
            return False
        raise AttributeError("Not a numerical term: " + str(other))

    def __ge__(self, other):
        if isinstance(other, Inf):
            if self._is_positive:
                return True
            else:
                return False
        if isinstance(other, numerical.Num):
            return self._is_positive
        elif isinstance(other, nan.Nan):
            return False
        raise AttributeError("Not a numerical term: " + str(other))

    def is_zero(self):
        return False

    def is_positive(self):
        return self._is_positive

    def is_negative(self):
        return not self._is_positive

    def is_inf(self):
        return True

    def is_inf_pos(self):
        return self._is_positive

    def is_inf_neg(self):
        return not self._is_positive

    def __hash__(self):
        if self.pos:
            return 17*hash("+INF")
        else:
            return 17*hash("-INF")

    def unpack(self):
        return str(self)



from aiddl_core.representation.str import Str


class StringConcatFunction:
    def __call__(self, x):
        s = '"'
        for e in x:
            s += e.string
        s += '"'
        return Str(s)



import aiddl_core.function.eval.higher_order
import aiddl_core.function.eval.misc
import aiddl_core.function.eval.python
import aiddl_core.function.eval.type
import aiddl_core.function.function
from aiddl_core.function.function_registry import FunctionRegistry
from aiddl_core.function.evaluator import Evaluator
from aiddl_core.representation.collection import Collection
from aiddl_core.representation.funref import FunRef
from aiddl_core.representation.inf import Inf
from aiddl_core.representation.int import Int
from aiddl_core.representation.keyval import KeyVal
from aiddl_core.representation.list import List
from aiddl_core.representation.num import Num
from aiddl_core.representation.rat import Rat
from aiddl_core.representation.real import Real
from aiddl_core.representation.entref import EntRef
from aiddl_core.representation.set import Set
from aiddl_core.representation.str import Str
from aiddl_core.representation.sym import Boolean, Sym
from aiddl_core.representation.tuple import Tuple
from aiddl_core.representation.var import Var

import aiddl_core.function.eval.numerical as numerical
import aiddl_core.function.eval.misc as misc
import aiddl_core.function.eval.type as types
import aiddl_core.function.eval.logic as logic
import aiddl_core.function.eval.collection as collection
import aiddl_core.function.eval.string as string

import aiddl_core.function as furi


def get_default_function_registry(container):
    freg = FunctionRegistry()

    evaluator = Evaluator(container)
    freg.add_function(furi.EVAL, evaluator)
    freg.add_function(furi.EVAL_REF, aiddl_core.function.eval.misc.EvalReferenceFunction(evaluator))
    freg.add_function(furi.EVAL_ALL_REFS,
                      aiddl_core.function.eval.misc.EvalAllReferencesFunction(evaluator))
    
    freg.add_function(furi.CALL, aiddl_core.function.eval.higher_order.CallFunction(freg))

    freg.add_function(furi.LAMBDA,
                      aiddl_core.function.function.LambdaEvaluator(freg))
    freg.add_function(furi.FUNCTION_LOADER,
                      aiddl_core.function.eval.python.PythonFunctionLoader(freg))
    freg.add_function(furi.FACTORY_LOADER,
                      aiddl_core.function.eval.python.PythonFunctionFactoryLoader(freg))

    freg.add_function(furi.CORE_LANG, lambda x: Sym("python"))

    # org.aiddl.eval.symbolic
    freg.add_function(furi.SYM_CONCAT, lambda x: x[0] + x[1])
    freg.add_function(furi.SYM_SPLIT, lambda x: List([Sym(e) for e in str(x).split(".")]))

    # org.aiddl.eval.string
    freg.add_function(furi.STR_CONCAT, string.StringConcatFunction())

    # org.aiddl.eval.numerical
    freg.add_function(furi.ADD, numerical.Addition())
    freg.add_function(furi.SUB, numerical.Subtraction())
    freg.add_function(furi.MULT, numerical.Multiplication())
    freg.add_function(furi.DIV, numerical.Division())
    freg.add_function(furi.MODULO, numerical.Modulo())
    freg.add_function(furi.GREATER_THAN, lambda x: Boolean.create(x.get(0) > x.get(1)))
    freg.add_function(furi.GREATER_THAN_EQ, lambda x: Boolean.create(x.get(0) >= x.get(1)))
    freg.add_function(furi.LESS_THAN, lambda x: Boolean.create(x.get(0) < x.get(1)))
    freg.add_function(furi.LESS_THAN_EQ, lambda x: Boolean.create(x.get(0) <= x.get(1)))
    freg.add_function(furi.IS_POS, lambda x: Boolean.create(x.is_positive()))
    freg.add_function(furi.IS_NEG, lambda x: Boolean.create(x.is_negative()))
    freg.add_function(furi.IS_ZERO, lambda x: Boolean.create(x.is_zero()))
    freg.add_function(furi.IS_INF, lambda x: Boolean.create(x.is_inf()))
    freg.add_function(furi.IS_INF_POS, lambda x: Boolean.create(x.is_inf_pos()))
    freg.add_function(furi.IS_INF_NEG, lambda x: Boolean.create(x.is_inf_neg()))
    freg.add_function(furi.IS_NAN, lambda x: Boolean.create(x.is_nan()))

    freg.add_function(furi.LET, misc.Let(evaluator))
    freg.add_function(furi.MAP, aiddl_core.function.eval.higher_order.Map(freg))
    freg.add_function(furi.FILTER, aiddl_core.function.eval.higher_order.Filter(freg))
    freg.add_function(furi.REDUCE, aiddl_core.function.eval.higher_order.Reduce(freg))

    freg.add_function(furi.ZIP, collection.Zip(evaluator))
    freg.add_function(furi.MATCH, misc.Match(evaluator))
    freg.add_function(furi.DOMAIN, misc.ExpandDomain(evaluator))

    freg.add_function(furi.NOT, lambda x: Boolean.create(not x.bool))
    freg.add_function(furi.AND, logic.And(evaluator))
    freg.add_function(furi.OR, logic.Or(evaluator))
    freg.add_function(furi.FORALL, logic.Forall(evaluator))
    freg.add_function(furi.EXISTS, logic.Exists(evaluator))
    freg.add_function(furi.IF, logic.If(evaluator))
    freg.add_function(furi.COND, logic.Cond(evaluator))

    freg.add_function(furi.IN, lambda x: Boolean(x[1].contains(x[0])))
    freg.add_function(furi.CONTAINS, lambda x: Boolean(x[0].contains(x[1])))
    freg.add_function(furi.CONTAINS_ALL, lambda x: Boolean(x[0].contains_all(x[1])))
    freg.add_function(furi.CONTAINS_ANY, collection.ContainsAny())
    freg.add_function(furi.ADD_COL, lambda x: x[0].add(x[1]))
    freg.add_function(furi.ADD_COL_ALL, lambda x: x[0].add_all(x[1]))
    freg.add_function(furi.REM_COL, lambda x: x[0].remove(x[1]))
    freg.add_function(furi.REM_COL_ALL, lambda x: x[0].remove_all(x[1]))
    freg.add_function(furi.UNION, collection.Union())
    freg.add_function(furi.CONCAT, collection.Concat())
    freg.add_function(furi.SUM, collection.Sum())
    freg.add_function(furi.CONTAINS_MATCH, collection.ContainsMatch())
    freg.add_function(furi.PUT_ALL, lambda x: x[0].put_all(x[1]))

    freg.add_function(furi.GET_IDX, lambda x: x[1][x[0].int_value])
    freg.add_function(furi.GET_KEY, lambda x: x[1][x[0]])
    freg.add_function(furi.CONTAINS_KEY, lambda x: Boolean.create(x[0].contains_key(x[1])))

    freg.add_function(furi.QUOTE, misc.Quote())
    freg.add_function(furi.EVAL_REF, misc.EvalRef(evaluator))
    freg.add_function(furi.MATCHES, misc.Matches())
    freg.add_function(furi.EQUALS, lambda x: Boolean.create(x[0] == x[1]))
    freg.add_function(furi.NOT_EQUALS, lambda x: Boolean.create(x[0] != x[1]))
    freg.add_function(furi.SUBSTITUTE, misc.SubstitutionFunction())
    freg.add_function(furi.UNIQUE_MAP, lambda x: Boolean.create(x.is_unique_map()) )

    freg.add_function(furi.FIRST, lambda x: x[0])
    freg.add_function(furi.LAST, lambda x: x[-1])
    freg.add_function(furi.KEY, lambda x: x.key)
    freg.add_function(furi.VALUE, lambda x: x.value)

    freg.add_function(furi.SIZE, lambda x: Int(len(x)))

    freg.add_function(furi.TYPE, types.EvalType(freg))
    freg.add_function(furi.TYPE_TERM, lambda x: Boolean.create(True))
    freg.add_function(furi.TYPE_NUMERICAL, lambda x: Boolean.create(isinstance(x, Num)))
    freg.add_function(furi.TYPE_INTEGER, lambda x: Boolean.create(isinstance(x, Int)))
    freg.add_function(furi.TYPE_RATIONAL, lambda x: Boolean.create(isinstance(x, Rat)))
    freg.add_function(furi.TYPE_REAL, lambda x: Boolean.create(isinstance(x, Real)))
    freg.add_function(furi.TYPE_INF, lambda x: Boolean.create(isinstance(x, Inf)))
    freg.add_function(furi.TYPE_SYMBOLIC, lambda x: Boolean.create(isinstance(x, Sym)))
    freg.add_function(furi.TYPE_BOOLEAN, lambda x: Boolean.create(isinstance(x, Boolean)))
    freg.add_function(furi.TYPE_STRING, lambda x: Boolean.create(isinstance(x, Str)))
    freg.add_function(furi.TYPE_VARIABLE, lambda x: Boolean.create(isinstance(x, Var)))
    freg.add_function(furi.TYPE_EREF, lambda x: Boolean.create(isinstance(x, EntRef)))
    freg.add_function(furi.TYPE_FREF, lambda x: Boolean.create(isinstance(x, FunRef)))
    freg.add_function(furi.TYPE_COLLECTION, lambda x: Boolean.create(isinstance(x, Collection)))
    freg.add_function(furi.TYPE_SET, lambda x: Boolean.create(isinstance(x, Set)))
    freg.add_function(furi.TYPE_LIST, lambda x: Boolean.create(isinstance(x, List)))
    freg.add_function(furi.TYPE_TUPLE, lambda x: Boolean.create(isinstance(x, Tuple)))
    freg.add_function(furi.TYPE_KEY_VALUE, lambda x: Boolean.create(isinstance(x, KeyVal)))

    freg.load_def(container)

    return freg



import aiddl_core.representation.term as term

class Substitution:

    def __init__(self, m=None):
        if m is None:
            self.m = {}
        else:
            self.m = m

    @staticmethod
    def from_term(t: 'Collection') -> 'Substitution':
        """ Create a substitution from a collection of key-value pairs

        :param t: collection term consisting of key-value pairs
        :return:
        """
        s = Substitution()
        for e in t:
            s.add(e.key, e.value)
        return s

    def add(self, t_from: 'Term', t_to: 'Term') -> bool:
        """ Attempt to add a new element to this substitution. If the replaced term t_from already exists no change
        will occur.

        :param t_from: term to be replaced by substitution
        :param t_to: replacement term
        :return: true if the added element was compatible with this substitution, false otherwise
        """
        assert isinstance(t_from, term.Term)
        assert isinstance(t_to, term.Term)
        if t_from not in self.m.keys():
            self.m[t_from] = t_to
            return True
        else:
            self.m[t_from] == t_to

    def add_substitution(self, sub: 'Substitution') -> bool:
        """ Add another substitution to this one, if it is compatible.

        If sub is incompatible self will not be changed.

        :param sub: a substitution
        :return: true if sub was compatible with self, false otherwise
        """
        for t_from in sub.m.keys():
            if t_from in self.m.keys():
                if self.m[t_from] != sub.m[t_from]:
                    return False
        for t_from in sub.m.keys():
            self.add(t_from, sub.m[t_from])
        return True

    def substitute(self, t: 'Term') -> 'Term':
        """ Substitute a term

        :param t: term to substitute
        :return: the substitution if one exists, t otherwise
        """
        if t in self.m.keys():
            return self.m[t]
        else:
            return t

    def copy(self) -> 'Substitution':
        """ Create a new substitution identical to this one

        :return: a cops of this substitution
        """
        c = Substitution()
        for k in self.m.keys():
            c.add(k, self.m[k])
        return c

    def __len__(self) -> int:
        """ Number of replacements in this substitution

        :return: number of replacements
        """
        return len(self.m)

    def __str__(self):
        s = "{"
        for k in self.m.keys():
            s += str(k)+":"+str(self.m[k])+" "
        s = s[0:-1] + "}"
        return s

    def __repr__(self):
        return str(self)



import aiddl_core.representation.term as term


class KeyVal(term.Term):
    __slots__ = ["_key", "_value"]

    def __init__(self, key, value):
        assert isinstance(key, term.Term)
        assert isinstance(value, term.Term)
        super(term.Term, self).__setattr__("_key", key)
        super(term.Term, self).__setattr__("_value", value)

    def substitute(self, s):
        return KeyVal(self._key.substitute(s), self._value.substitute(s))

    def match(self, other):
        if isinstance(other, KeyVal):
            sub_key = self._key.match(other._key)
            if sub_key is None:
                return None
            sub_value = self._value.match(other._value)
            if sub_value is None:
                return None
            if not sub_key.add_substitution(sub_value):
                return None
            return sub_key
        return None

    @property
    def key(self):
        return self._key

    @property
    def value(self):
        return self._value

    def resolve(self, container):
        return KeyVal(self._key.resolve(container),
                      self._value.resolve(container))

    def __repr__(self):
        return str(self._key)+":"+str(self._value)

    def __str__(self):
        return str(self._key)+":"+str(self._value)

    def __eq__(self, other):
        if isinstance(other, KeyVal):
            return self._value == other._value and self._key == other._key
        return False

    def __ne__(self, other):
        return not (self == other)

    def __hash__(self):
        return 19*hash(self._key) + 3*hash(self._value)

    def unpack(self):
        return (self._key.unpack(), self._value.unpack())



from aiddl_core.container.entry import Entry
from aiddl_core.representation.sym import Sym
from aiddl_core.representation.substitution import Substitution

MOD = Sym("#mod")


class Module:
    def __init__(self, module_uri: Sym):
        """ Create a new module

        :param module_uri: name of the module
        """
        self._module_uri = module_uri
        self._data = {}

    @property
    def name(self) -> Sym:
        """ Name of the module

        :return: name of the module
        """
        return self._module_uri

    def put_entry(self, e):
        """ Add an entry to this module.

        Existing entries with the same name will be overwritten.

        :param e: added entry
        """
        self._data[e.name] = e

    def remove_entry(self, e: Entry):
        """ Remove an entry with the same name as e.

        :param e: entry to remove
        """
        del self._data[e.name]

    def get_entries(self):
        """ Get all entries from this module.

        :return: list of entries
        """
        return self._data.values()

    def substitute(self, s: Substitution):
        """ Apply a substitution to all entries in this module

        :param s: substitution to apply
        """
        subbed_entries = []
        for e in self._data.values():
            subbed_entries.append(e.substitute(s))
        for e in subbed_entries:
            self.put_entry(e)




import aiddl_core.representation.term as term
import aiddl_core.representation.num as numerical
import aiddl_core.representation.int as integer
import aiddl_core.representation.rat as rational
import aiddl_core.representation.inf as infinity
import aiddl_core.representation.nan as nan

from math import floor


class Real(numerical.Num):
    __slots__ = ["_value"]

    def __init__(self, value):
        super(term.Term, self).__setattr__("_value", value)

    @property
    def real_value(self):
        """ Return this integer as a python float object

        :return: float value
        """
        return self._value

    @property
    def int_value(self):
        """ Return this integer as a python int object

        :return: int value
        """
        return int(self._value)

    def __add__(self, other):
        if isinstance(other, Real) or isinstance(other, integer.Int):
            return Real(self._value + other._value)
        elif isinstance(other, rational.Rat):
            return Real(self._value + float(other._p)/float(other._q))
        elif isinstance(other, infinity.Inf):
            return other
        elif isinstance(other, nan.NaN):
            return other
        raise AttributeError("Not a numerical term: " + str(other))

    def __sub__(self, other):
        if isinstance(other, Real) or isinstance(other, integer.Int):
            return Real(self._value - other._value)
        elif isinstance(other, rational.Rat):
            return Real(self._value - float(other._p)/float(other._q))
        elif isinstance(other, infinity.Inf):
            return infinity.Inf(not other._is_positive)
        elif isinstance(other, nan.NaN):
            return other
        raise AttributeError("Not a numerical term: " + str(other))

    def __mul__(self, other):
        if isinstance(other, Real) or isinstance(other, integer.Int):
            return Real(self._value * other._value)
        elif isinstance(other, rational.Rat):
            return Real(self._value * float(other._p)/float(other._q))
        elif isinstance(other, infinity.Inf):
            if self == integer.Int(0):
                return nan.NaN()
            else:
                return infinity.Inf(other._is_positive
                                    == (self._value > 0))
        elif isinstance(other, nan.NaN):
            return other
        raise AttributeError("Not a numerical term: " + str(other))

    def __truediv__(self, other):
        if isinstance(other, Real) or isinstance(other, integer.Int):
            return Real(self._value / other._value)
        elif isinstance(other, rational.Rat):
            return Real(self._value / (other._p/other._q))
        elif isinstance(other, infinity.Inf):
            return integer.Int(0)
        elif isinstance(other, nan.NaN):
            return other
        raise AttributeError("Not a numerical term: " + str(other))

    def __floordiv__(self, other):
        if isinstance(other, Real) or isinstance(other, integer.Int):
            return integer.Int(floor(self._value / other._value))
        elif isinstance(other, rational.Rat):
            return integer.Int(int(floor(self._value /
                                         other._p /
                                         other._q)))
        elif isinstance(other, infinity.Inf):
            return integer.Int(0)
        elif isinstance(other, nan.NaN):
            return other
        raise AttributeError("Not a numerical term: " + str(other))

    def __repr__(self):
        return str(self._value)
    
    def __str__(self):
        return str(self._value)

    def __eq__(self, other):
        if isinstance(other, Real) or isinstance(other, integer.Int):
            return self._value == other._value
        elif isinstance(other, rational.Rat):
            return self._value == float(other._p)/float(other._q)
        else:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __lt__(self, other):
        if isinstance(other, Real) or isinstance(other, integer.Int):
            return self._value < other._value
        elif isinstance(other, rational.Rat):
            return self._value < float(other._p)/float(other._q)
        elif isinstance(other, infinity.Inf):
            return other._is_positive
        elif isinstance(other, nan.NaN):
            return False
        raise AttributeError("Not a numerical term: " + str(other))

    def __le__(self, other):
        if isinstance(other, Real) or isinstance(other, integer.Int):
            return self._value <= other._value
        elif isinstance(other, rational.Rat):
            return self._value <= float(other._p)/float(other._q)
        elif isinstance(other, infinity.Inf):
            return other._is_positive
        elif isinstance(other, nan.NaN):
            return False
        raise AttributeError("Not a numerical term: " + str(other))

    def __gt__(self, other):
        if isinstance(other, Real) or isinstance(other, integer.Int):
            return self._value > other._value
        elif isinstance(other, rational.Rat):
            return self._value > float(other._p)/float(other._q)
        elif isinstance(other, infinity.Inf):
            return not other._is_positive
        elif isinstance(other, nan.NaN):
            return False
        raise AttributeError("Not a numerical term: " + str(other))

    def __ge__(self, other):
        if isinstance(other, Real) or isinstance(other, integer.Int):
            return self._value >= other._value
        elif isinstance(other, rational.Rat):
            return self._value >= float(other._p)/float(other._q)
        elif isinstance(other, infinity.Inf):
            return not other._is_positive
        elif isinstance(other, nan.NaN):
            return False
        raise AttributeError("Not a numerical term: " + str(other))

    def is_zero(self):
        return self._value == 0.0

    def is_positive(self):
        return self._value > 0.0

    def is_negative(self):
        return self._value < 0.0

    def __hash__(self):
        return 3*hash(self._value)

    def unpack(self):
        return self._value



from aiddl_core.representation.inf import Inf
from aiddl_core.representation.int import Int
from aiddl_core.representation.keyval import KeyVal
from aiddl_core.representation.list import List
from aiddl_core.representation.num import Num
from aiddl_core.representation.set import Set
from aiddl_core.representation.sym import Sym, Boolean
from aiddl_core.representation.collection import Collection
from aiddl_core.representation.substitution import Substitution
from aiddl_core.representation.funref import FunRef
from aiddl_core.representation.sym import TRUE
from aiddl_core.representation.sym import FALSE

import aiddl_core.function as furi
from aiddl_core.representation.tuple import Tuple
from aiddl_core.util.logger import Logger

SELF = Sym("#self")
SELF_ALT = Sym("#arg")


class EvalType:
    def __init__(self, freg):
        self.evaluator = freg.get_function(furi.EVAL)
        self.freg = freg

    def __call__(self, args):
        x = args[0]
        type_def = args[1]
        if isinstance(type_def, Collection):
            for t in type_def:
                if t == FALSE:
                    return t
                answer = self.check_type(t, x)
                if answer == TRUE:
                    return answer
            return FALSE

        if type_def == FALSE:
            return FALSE
        return self.check_type(type_def, x)

    def check_type(self, type_def, x):
        if isinstance(type_def, Sym):
            t_check = self.fReg.get_function(type_def)
            print("[W] Symbolic function reference:", type_def, "for", x)
            return t_check(x)
        elif isinstance(type_def, FunRef):
            if type_def.function is not None:
                return type_def.function(x)
            else:
                print("Function not found:", type_def)
        print("[W] Not a function reference:", type_def, "for:", x, "it has type:", type(type_def))
        self_sub = Substitution()
        self_sub.add(SELF, x)
        self_sub.add(SELF_ALT, x)
        return self.evaluator(type_def.substitute(self_sub))


class TypeCheckFunction:
    def __init__(self, type_def, evaluator):
        self.type_def = type_def
        self.evaluator = evaluator

    def __call__(self, term):
        return Boolean.create(self.check(self.type_def, term))

    def check(self, type_def, term):
        r = False
        if isinstance(type_def, Tuple):
            type_class = type_def[0]

            if type_class == Sym("basic-type"):
                e = Tuple([type_def[1], term])
                r = self.evaluator(e).bool
            elif type_class == Sym("org.aiddl.type.set-of"):
                sub_type = type_def.get(1)
                if isinstance(term, Set):
                    r = True
                    Logger.inc_depth()
                    for e in term:
                        if not self.check(sub_type, e):
                            #print("---> FAIL")
                            r = False
                            break
                    Logger.dec_depth()
                else:
                    r = False
            elif type_class == Sym("org.aiddl.type.list-of"):
                subType = type_def.get(1)
                if isinstance(term, List):
                    r = True
                    Logger.inc_depth()
                    for e in term:
                        if not self.check(subType, e):
                            r = False
                            break
                    Logger.dec_depth()
                else:
                    r = False
            elif type_class == Sym("org.aiddl.type.collection-of"):
                subType = type_def.get(1)
                if isinstance(term, Collection):
                    r = True
                    Logger.inc_depth()
                    for e in term:
                        if not self.check(subType, e):
                            r = False
                            break
                    Logger.dec_depth()
                else:
                    r = False
            elif type_class == Sym("org.aiddl.type.tuple.signed"):
                if isinstance(term, Tuple):
                    signature = type_def.get(1)
                    min = type_def.get_or_default(Sym("min"), Int(len(signature)))
                    max = type_def.get_or_default(Sym("max"), Int(len(signature)))
                    repeat = type_def.get_or_default(Sym("repeat"), Int(1))
                    tSize = Int(len(term))
                    repeat_start_idx = len(signature) - repeat.int_value

                    if tSize >= min and tSize <= max:
                        Logger.inc_depth()
                        r = True
                        for i in range(0, len(signature)):
                            sig_idx = i
                            if i >= signature.size():
                                sig_idx = repeat_start_idx + \
                                    (i - signature.size()) % repeat.int_value
                                if not self.check(signature.get(sig_idx), term.get(i)):
                                    r = False
                                    break
                        Logger.dec_depth()
                    else:
                        r = False
                else:
                    r = False
            elif type_class == Sym("org.aiddl.type.dictionary"):
                keyTypeCol = type_def.get(1)
                r = True
                Logger.inc_depth()
                for kvp in keyTypeCol:
                    e = term.get(kvp.key)
                    if e is None:
                        r = False
                        break
                    if not self.check(kvp.value, e):
                        r = False
                        break

                optional = type_def.get_or_default(Sym("optional"), Set([]))
                for kvp in optional:
                    e = term.get(kvp.key)
                    if e is not None:
                        if not self.check(kvp.value, e):
                            r = False
                            break

                Logger.dec_depth()
            elif type_class == Sym("org.aiddl.type.matrix"):
                if (isinstance(term, Tuple) or isinstance(term, List)) and len(term) > 0:
                    m = type_def.get_or_default(Sym("m"), Int(len(term))).unpack()
                    n = type_def.get_or_default(Sym("n"), Int(len(term[0]))).unpack()
                    cell_type = type_def.get(Sym("cell-type"))
                    row_types = type_def.get(Sym("row-type"))
                    col_types = type_def.get(Sym("col-type"))
                    r = True
                    if len(term) != m:
                        r = False
                    else:
                        for i in range(m):
                            if not r or len(term[i]) != n:
                                r = False
                                break
                            for j in range(m):
                                cell_okay = cell_type is not None and self.check(cell_type, term[i][j])
                                row_okay = row_types is not None and self.check(row_types[i], term[i][j])
                                col_okay = col_types is not None and self.check(col_types[j], term[i][j])
                                if not (cell_okay and row_okay and col_okay):
                                    r = False
                                    break
                else:
                    r = False

            elif type_class == Sym("org.aiddl.type.enum"):
                r = term in type_def.get(1)
            elif type_class == Sym("org.aiddl.type.range"):
                min_val = type_def.get_or_default(Sym("min"), Inf.neg())
                max_val = type_def.get_or_default(Sym("max"), Inf.pos())
                r = False
                if isinstance(term, Num):
                    if min_val <= term <= max_val:
                        r = True
            elif type_class == Sym("org.aiddl.type.typed-key-value"):
                r = False
                if isinstance(term, KeyVal):
                    if self.check(type_def[1].key, term.key) and self.check(type_def[1].value, term.value):
                        r = True
            elif type_class == Sym("org.aiddl.type.union"):
                r = False
                for choice in type_def[1]:
                    if self.check(choice, term):
                        r = True
                        break
            elif type_class == Sym("org.aiddl.type.intersection"):
                r = True
                for choice in type_def[1]:
                    if not self.check(choice, term):
                        r = False
                        break
            else:
                raise ValueError("#type expression not supported:", type_def)
        elif isinstance(type_def, Sym):
            e = Tuple([type_def, term])
            r = self.evaluator(e).bool
        elif isinstance(type_def, FunRef):
            r = type_def(term).bool
        else:
            r = False
            raise ValueError("#type expression not supported (%s): %s" % (str(type(type_def)), str(type_def)))

        if r and isinstance(type_def, Tuple):
            constraint = type_def[Sym("constraint")]
            if constraint is not None and isinstance(constraint, FunRef):
                r = constraint.function(term).bool
                # if not r:
                #     if eval.get_verbosity() >= 1:
                #         Logger.incDepth()
                #         Logger.msg("TypeCheck", t, " does not satisfy " + constraint)
                #         Logger.decDepth()

        # if not r:
        #     Logger.msg("TypeCheck", str(term) + " !!  " + str(type_def))
        return r


class GenericTypeConstructor:
    def __init__(self, uri_base, args, type_def, evaluator, freg):
        self.next_free_id = 0
        self.uri_base = uri_base
        self.args = args
        self.type_def = type_def
        self.evaluator = evaluator
        self.freg = freg

    def __call__(self, x):
        self.next_free_id += 1
        s = self.args.match(x)
        type_def = self.type_def.substitute(s)
        uri = self.uri_base + Sym("n%d" % self.next_free_id)
        print(type_def)
        type_fun = TypeCheckFunction(type_def, self.evaluator)
        self.freg.add_function(uri, type_fun)
        return FunRef(uri, self.freg)



from .function import FunctionMixin
from .function import LazyFunctionMixin
from .function import InitializableMixin
from .function import ConfigurableMixin
from .function import InterfaceImplementationMixin
from .evaluator import Evaluator
from .function_registry import FunctionRegistry

from aiddl_core.representation.sym import Sym

EVAL = Sym("org.aiddl.eval")
EVAL_REF = Sym("org.aiddl.eval.eval-ref")
EVAL_ALL_REFS = Sym("org.aiddl.eval.eval-all-refs")

FUNCTION_LOADER = Sym("org.aiddl.eval.load-function")
FACTORY_LOADER = Sym("org.aiddl.eval.load-function-factory")
CORE_LANG = Sym("org.aiddl.eval.core-lang")

# org.aiddl.eval

CALL = Sym("org.aiddl.eval.call")
LAMBDA = Sym("org.aiddl.eval.lambda")
QUOTE = Sym("org.aiddl.eval.quote")
TYPE = Sym("org.aiddl.eval.type")
ZIP = Sym("org.aiddl.eval.zip")
MAP = Sym("org.aiddl.eval.map")
FILTER = Sym("org.aiddl.eval.filter")
REDUCE = Sym("org.aiddl.eval.reduce")
DOMAIN = Sym("org.aiddl.eval.domain")
EVAL_REF = Sym("org.aiddl.eval.eval-ref")
MATCHES = Sym("org.aiddl.eval.matches")
EQUALS = Sym("org.aiddl.eval.equals")
NOT_EQUALS = Sym("org.aiddl.eval.not-equals")
SUBSTITUTE = Sym("org.aiddl.eval.substitute")
GET_MATCHING_ENTRIES = Sym("org.aiddl.eval.get-matching-entries")

FIRST = Sym("org.aiddl.eval.first")
LAST = Sym("org.aiddl.eval.last")
KEY = Sym("org.aiddl.eval.key")
VALUE = Sym("org.aiddl.eval.value")
SIZE = Sym("org.aiddl.eval.size")
GET_IDX = Sym("org.aiddl.eval.get-idx")
GET_KEY = Sym("org.aiddl.eval.get-key")
CONTAINS_KEY = Sym("org.aiddl.eval.contains-key")
IF = Sym("org.aiddl.eval.if")
COND = Sym("org.aiddl.eval.cond")
MATCH = Sym("org.aiddl.eval.match")
LET = Sym("org.aiddl.eval.let")
UNIQUE_MAP = Sym("org.aiddl.eval.is-unique-map")

# org.aiddl.eval.symbolic

SYM_CONCAT = Sym("org.aiddl.eval.symbolic.concat")
SYM_SPLIT = Sym("org.aiddl.eval.symbolic.split")

# org.aiddl.eval.string

STR_CONCAT = Sym("org.aiddl.eval.string.concat")
STR_MOD_NAME = Sym("org.aiddl.eval.string.module-folder")

# org.aiddl.eval.collection

IN = Sym("org.aiddl.eval.collection.in")
CONTAINS = Sym("org.aiddl.eval.collection.contains")
CONTAINS_ALL = Sym("org.aiddl.eval.collection.contains-all")
CONTAINS_ANY = Sym("org.aiddl.eval.collection.contains-any")
ADD_COL = Sym("org.aiddl.eval.collection.add-element")
ADD_COL_ALL = Sym("org.aiddl.eval.collection.add-all")
REM_COL = Sym("org.aiddl.eval.collection.remove")
REM_COL_ALL = Sym("org.aiddl.eval.collection.remove-all")
CONTAINS_MATCH = Sym("org.aiddl.eval.collection.contains-match")
PUT_ALL = Sym("org.aiddl.eval.collection.put-all")
SUM = Sym("org.aiddl.eval.collection.sum")

# org.aiddl.eval.set

UNION = Sym("org.aiddl.eval.set.union")

# org.aiddl.eval.list

CONCAT = Sym("org.aiddl.eval.list.concat")

# org.aiddl.eval.numeric

ADD = Sym("org.aiddl.eval.numerical.add")
SUB = Sym("org.aiddl.eval.numerical.sub")
MULT = Sym("org.aiddl.eval.numerical.mult")
DIV = Sym("org.aiddl.eval.numerical.div")
MODULO = Sym("org.aiddl.eval.numerical.modulo")
GREATER_THAN = Sym("org.aiddl.eval.numerical.greater-than")
GREATER_THAN_EQ = Sym("org.aiddl.eval.numerical.greater-than-eq")
LESS_THAN = Sym("org.aiddl.eval.numerical.less-than")
LESS_THAN_EQ = Sym("org.aiddl.eval.numerical.less-than-eq")
IS_POS = Sym("org.aiddl.eval.numerical.is-positive")
IS_NEG = Sym("org.aiddl.eval.numerical.is-negative")
IS_ZERO = Sym("org.aiddl.eval.numerical.is-zero")
IS_INF = Sym("org.aiddl.eval.numerical.is-infinite")
IS_INF_POS = Sym("org.aiddl.eval.numerical.is-infinite-positive")
IS_INF_NEG = Sym("org.aiddl.eval.numerical.is-infinite-negative")
IS_NAN = Sym("org.aiddl.eval.numerical.is-nan")

# org.aiddl.eval.random

UNIFORM_RANDOM = Sym("org.aiddl.eval.random.uniform.real")
UNIFORM_RANDOM_ELEMENT = Sym("org.aiddl.eval.random.uniform.element")
UNIFORM_RANDOM_INTEGER = Sym("org.aiddl.eval.random.uniform.integer")
NORMAL_RANDOM = Sym("org.aiddl.eval.random.normal")


# org.aiddl.eval.logic

AND = Sym("org.aiddl.eval.logic.and")
OR = Sym("org.aiddl.eval.logic.or")
NOT = Sym("org.aiddl.eval.logic.not")
EXISTS = Sym("org.aiddl.eval.logic.exists")
FORALL = Sym("org.aiddl.eval.logic.forall")

# org.aiddl.type

TYPE_TERM = Sym("org.aiddl.type.term")
TYPE_NUMERICAL = Sym("org.aiddl.type.term.numerical")
TYPE_INTEGER = Sym("org.aiddl.type.term.numerical.integer")
TYPE_RATIONAL = Sym("org.aiddl.type.term.numerical.rational")
TYPE_REAL = Sym("org.aiddl.type.term.numerical.real")
TYPE_INF = Sym("org.aiddl.type.term.numerical.infinity")
TYPE_SYMBOLIC = Sym("org.aiddl.type.term.symbolic")
TYPE_BOOLEAN = Sym("org.aiddl.type.term.symbolic.boolean")
TYPE_STRING = Sym("org.aiddl.type.term.string")
TYPE_VARIABLE = Sym("org.aiddl.type.term.variable")
TYPE_EREF = Sym("org.aiddl.type.term.ref.entry")
TYPE_FREF = Sym("org.aiddl.type.term.ref.function")
TYPE_COLLECTION = Sym("org.aiddl.type.term.collection")
TYPE_SET = Sym("org.aiddl.type.term.collection.set")
TYPE_LIST = Sym("org.aiddl.type.term.collection.list")
TYPE_TUPLE = Sym("org.aiddl.type.term.tuple")
TYPE_KEY_VALUE = Sym("org.aiddl.type.term.key-value")




import aiddl_core.representation.term as term
import aiddl_core.representation.num as numerical
import aiddl_core.representation.rat as rational
import aiddl_core.representation.real as real
import aiddl_core.representation.inf as infinity
import aiddl_core.representation.nan as nan

from math import floor


class Int(numerical.Num):
    __slots__ = ["_value"]

    def __init__(self, value):
        super(term.Term, self).__setattr__("_value", value)

    @property
    def int_value(self):
        """ Return this integer as a python int object

        :return: int value
        """
        return self._value

    @property
    def real_value(self):
        """ Return this integer as a float

        :return: real value
        """
        return float(self._value)

    def __add__(self, other):
        if isinstance(other, Int):
            return Int(self._value + other._value)
        elif isinstance(other, rational.Rat):
            return rational.Rat(self._value * other._q + other._p, other._q)
        elif isinstance(other, real.Real):
            return real.Real(self._value + other._value)
        elif isinstance(other, infinity.Inf):
            return other
        elif isinstance(other, nan.NaN):
            return other
        raise AttributeError("Not a numerical term: " + str(other) + " type: " + str(type(other)) )

    def __sub__(self, other):
        if isinstance(other, Int):
            return Int(self._value - other._value)
        elif isinstance(other, rational.Rat):
            return rational.Rat(self._value * other._q - other._p,
                                other._q)
        elif isinstance(other, real.Real):
            return real.Real(self._value - other._value)
        elif isinstance(other, infinity.Inf):
            return infinity.Inf(not other._is_positive)
        elif isinstance(other, nan.NaN):
            return other
        raise AttributeError("Not a numerical term: " + str(other))

    def __mul__(self, other):
        if isinstance(other, Int):
            return Int(self._value * other._value)
        elif isinstance(other, rational.Rat):
            return rational.Rat(self._value * other._p, other._q)
        elif isinstance(other, real.Real):
            return real.Real(self._value * other._value)
        elif isinstance(other, infinity.Inf):
            if self._value == 0:
                return nan.NaN()
            else:
                return infinity.Inf(other._is_positive ==
                                    (self._value > 0))
        elif isinstance(other, nan.NaN):
            return other
        raise AttributeError("Not a numerical term: " + str(other))

    def __truediv__(self, other):
        if isinstance(other, numerical.Num) and other.is_zero():
            return nan.NaN()
        if isinstance(other, Int):
            return real.Real(self._value / other._value)
        elif isinstance(other, rational.Rat):
            return rational.Rat(self._value * other._q, other._p)
        elif isinstance(other, real.Real):
            return rational.Rat(self._value, other._value)
        elif isinstance(other, infinity.Inf):
            return Int(0)
        elif isinstance(other, nan.NaN):
            return other
        raise AttributeError("Not a numerical term: " + str(other))

    def __floordiv__(self, other):
        if isinstance(other, numerical.Num) and other.is_zero():
            return nan.NaN()
        if isinstance(other, Int):
            return Int(self._value // other._value)
        elif isinstance(other, rational.Rat):
            return Int((self._value * other._q) // other._p)
        elif isinstance(other, real.Real):
            return Int(int(floor(self._value / other._value)))
        elif isinstance(other, infinity.Inf):
            return Int(0)
        elif isinstance(other, nan.NaN):
            return other
        raise AttributeError("Not a numerical term: " + str(other))

    def __str__(self):
        return str(self._value)

    def __repr__(self):
        return str(self._value)

    def __eq__(self, other):
        if isinstance(other, Int):
            return self._value == other._value
        elif isinstance(other, rational.Rat):
            return self._value*other._q == other._p
        elif isinstance(other, real.Real):
            return float(self._value) == other._value
        elif isinstance(other, infinity.Inf):
            return False
        else:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __lt__(self, other):
        if isinstance(other, Int):
            return self._value < other._value
        elif isinstance(other, rational.Rat):
            return self._value*other._q < other._p
        elif isinstance(other, real.Real):
            return float(self._value) < other._value
        elif isinstance(other, infinity.Inf):
            return other._is_positive
        elif isinstance(other, nan.NaN):
            return False
        raise AttributeError("Not a numerical term: " + str(other))

    def __le__(self, other):
        if isinstance(other, Int):
            return self._value <= other._value
        elif isinstance(other, rational.Rat):
            return self._value*other._q <= other._p
        elif isinstance(other, real.Real):
            return float(self._value) <= other._value
        elif isinstance(other, infinity.Inf):
            return other._is_positive
        elif isinstance(other, nan.NaN):
            return False
        raise AttributeError("Not a numerical term: " + str(other))

    def __gt__(self, other):
        if isinstance(other, Int):
            return self._value > other._value
        elif isinstance(other, rational.Rat):
            return self._value*other._q > other._p
        elif isinstance(other, real.Real):
            return float(self._value) > other._value
        elif isinstance(other, infinity.Inf):
            return not other._is_positive
        elif isinstance(other, nan.NaN):
            return False
        raise AttributeError("Not a numerical term: " + str(other))

    def __ge__(self, other):
        if isinstance(other, Int):
            return self._value >= other._value
        elif isinstance(other, rational.Rat):
            return self._value*other._q >= other._p
        elif isinstance(other, real.Real):
            return float(self._value) >= other._value
        elif isinstance(other, infinity.Inf):
            return not other._is_positive
        elif isinstance(other, nan.NaN):
            return False
        raise AttributeError("Not a numerical term: " + str(other))

    def is_zero(self):
        return self._value == 0

    def is_positive(self):
        return self._value > 0

    def is_negative(self):
        return self._value < 0

    def __hash__(self):
        return hash(self._value)

    def unpack(self):
        return self._value



from aiddl_core.representation.set import Set
from aiddl_core.representation.list import List
from aiddl_core.representation.tuple import Tuple
from aiddl_core.representation.keyval import KeyVal


class Logger:
    depth = 0

    def inc_depth():
        Logger.depth += 1

    def dec_depth():
        Logger.depth -= 1

    def tabbing():
        return "  "*Logger.depth

    def msg(source, msg):
        print("%s[%s] %s" % (Logger.tabbing(), source, str(msg)))

    def pretty_print(t, depth):
        base = str(t)
        if len(base) > 80:
            s = ""
            if isinstance(t, Set):
                s += Logger.simple_tabbing(depth)
                s += "{\n"
                for e in t:
                    s += Logger.pretty_print(e, depth + 1)
                    s += "\n"
                s += Logger.simple_tabbing(depth)
                s += "}\n"
            elif isinstance(t, List):
                s += Logger.simple_tabbing(depth)
                s += "[\n"
                for e in t:
                    s += Logger.pretty_print(e, depth + 1)
                    s += "\n"
                s += Logger.simple_tabbing(depth)
                s += "]\n"
            elif isinstance(t, Tuple):
                s += Logger.simple_tabbing(depth)
                s += "(\n"
                for e in t:
                    s += Logger.pretty_print(e, depth + 1)
                    s += "\n"
                s += Logger.simple_tabbing(depth)
                s += ")\n"
            elif isinstance(t, KeyVal):
                s += Logger.pretty_print(t.key, depth)
                s += ":\n"
                s += Logger.pretty_print(t.value, depth+1)
            else:
                s += Logger.simple_tabbing(depth)
                s += base
            return s
        return Logger.simple_tabbing(depth) + base

    def simple_tabbing(n):
        return "  " * n



import aiddl_core.representation.term as term


class Str(term.Term):
    __slots__ = ["_value"]

    def __init__(self, value):
        if value[0] != '"':
            value = '"' + value
        if value[-1] != '"' or value[-2] == '\\':
            value = value + '"'  

        super(term.Term, self).__setattr__("_value", value)

    @property
    def string(self):
        return self._value[1:-1] # .replace('\\"', '"')

    def __str__(self):
        return str(self._value)

    def __repr__(self):
        return str(self._value)

    def __eq__(self, other):
        if isinstance(other, Str):
            return self._value == other._value
        return False

    def __ne__(self, other):
        return not self == other

    def __hash__(self):
        return 17*hash(self._value)

    def unpack(self):
        return self._value



from aiddl_core.representation.sym import Sym
from aiddl_core.representation.var import Var
from aiddl_core.representation.tuple import Tuple
from aiddl_core.representation.list import List

from aiddl_core.function.eval.type import TypeCheckFunction, GenericTypeConstructor

from aiddl_core.function.function import InterfaceImplementationMixin, NamedFunction, LambdaFunction
import aiddl_core.function as fun_uri

DEF = Sym("#def")


class FunctionRegistry:
    def __init__(self):
        self.functions = {}
        self.interfaces = {}
        self.interface_implementations = {}

    def add_function(self, name, f):
        self.functions[name] = f
        for k in self.interface_implementations.keys():
            self.interface_implementations[k] = filter(
                name.__ne__,
                self.interface_implementations[k])
        if isinstance(f, InterfaceImplementationMixin):
            uri = f.get_interface_uri()
            if uri not in self.interface_implementations.keys():
                self.interface_implementations[uri] = []
            self.interface_implementations[uri].append(name)

    def get_function(self, name):
        if isinstance(name, Tuple) \
           and len(name) == 3 \
           and name[0] == fun_uri.LAMBDA:
            return self.lambda_factory(name)
        if name in self.functions.keys():
            return self.functions[name]
        return None

    def get_function_or_panic(self, name):
        if isinstance(name, Tuple) \
           and len(name) == 3 \
           and name[0] == fun_uri.LAMBDA:
            return self.lambda_factory(name)
        if name in self.functions.keys():
            return self.functions[name]
        raise ValueError("Function not registered:", str(name))

    def get_function_or_default(self, name, f_def):
        if isinstance(name, Tuple) \
           and len(name) == 3 \
           and name[0] == fun_uri.LAMBDA:
            return self.lambda_factory(name)
        if name in self.functions.keys():
            return self.functions[name]
        raise ValueError(f_def)

    def get_interface_implementations(self, uri):
        if uri in self.interface_implementations:
            return List(self.interface_implementations[uri])
        return List()

    def lambda_factory(self, n):
        x = n.get(1)
        f = n.get(2)
        e = self.functions[fun_uri.EVAL]
        return LambdaFunction(x, f, e)

    def has_function(self, name):
        return name in self.functions.keys()

    def get_registered_names(self):
        return list(self.functions.keys())

    def load_container_interfaces(self, C):
        evaluator = self.get_function(fun_uri.EVAL)
        for m in C.get_module_names():
            for e in C.get_matching_entries(m, Sym("#interface"), Var()):
                uri = evaluator(e.value[Sym("uri")])
                interface_term = evaluator(e.value)
                self.interfaces[uri] = interface_term

    # def get_function_list(self, m, C):
    #     L = []
    #     for e in C.get_matching_entries(m, Symbolic("#functions"), Variable()):
    #         for t in e.get_value():
    #             L.append(t)
    #     return List(L)
    #
    # def load_req_python_functions(self, C):
    #     for m in C.get_module_names():
    #         functions = self.get_function_list(m, C) # C.get_entry(Symbolic("functions"), module=m)
    #         # print("Functions:", functions)
    #         if len(functions) > 0:
    #             missing = False
    #             for f in functions:
    #                 if not self.has_function(f[0]):
    #                     missing = True
    #                     break
    #             if missing:
    #                 loader_mod = m + Symbolic("python")
    #                 if parser.is_known_module(loader_mod):
    #                     lu = parser.get_mod_file_lookup(parser.collect_aiddl_paths([]))
    #
    #                     evalutaor = self.get_function(EVAL)
    #                     parser.parse_internal(lu[loader_mod], C, self, ".")
    #
    #                     for e in C.get_matching_entries(loader_mod, Symbolic("#on-load"), Variable()):
    #                         load = e.get_value()
    #
    #                         if isinstance(load, Tuple):
    #                             evalutaor(load)
    #                         else:
    #                             for call in load:
    #                                 evalutaor(call)
    #
    #                     # load_request = C.get_entry(Symbolic("load"), module=loader_mod)
    #                     # if load_request is not None:
    #                     #     rHandler = RequestHandler(C, self)
    #                     #     # rHandler.verbose = True
    #                     #     rHandler.satisfy_request(load_request.get_value(), Symbolic("NIL"))
    #                 # else:
    #                 #     print("Could not find loader module:", loader_mod)
    #                 for f in functions:
    #                     if not self.has_function(f[0]):
    #                         print("[Warning]", f[0], ": Missing python implementation")

    def load_type_functions(self, C):
        evaluator = self.get_function(fun_uri.EVAL)
        for m in C.get_module_names():
            for e in C.get_matching_entries(m, Sym("#type"), Var()):
                if isinstance(e.name, Sym):
                    uri = m + e.name
                    evaluator.set_follow_references(True)
                    #evaluator.set_verbose(True)
                    type_def = evaluator(e.value)
                    evaluator.set_follow_references(False)
                    #evaluator.set_verbose(False)
                    type_fun = TypeCheckFunction(type_def, evaluator)
                    self.add_function(uri, type_fun)

                    interface_term = evaluator(e.value)
                    self.interfaces[uri] = interface_term
                    # print("Loaded type:", uri, "with def", type_def)
                elif isinstance(e.name, Tuple):
                    base_uri = m + e.name[0]
                    evaluator.set_follow_references(True)
                    type_def = evaluator(e.value)
                    evaluator.set_follow_references(False)
                    arg_list = []
                    for i in range(1, len(e.name)):
                        arg_list.append(e.name[i])
                    if len(arg_list) == 1:
                        gen_args = arg_list[0]
                    else:
                        gen_args = Tuple(arg_list)
                    type_fun = GenericTypeConstructor(base_uri, gen_args, type_def, evaluator, self)
                    self.add_function(base_uri, type_fun)


    def load_def(self, C):
        for m in C.get_module_names():
            for e in C.get_matching_entries(m, DEF, Var()):
                if isinstance(e.name, Sym):
                    uri = m + e.name
                    f = NamedFunction(uri,
                                      e.value,
                                      self.functions[fun_uri.EVAL])
                    # print("Loading:", uri)
                else:
                    uri = m + e.name[0]

                    args = None
                    if len(e.name) == 2:
                        args = e.name[1]
                    else:
                        arg_list = []
                        for i in range(1, len(e.name)):
                            arg_list.append(e.name[i])
                        args = Tuple(arg_list)
                    # print("Loading:", uri)
                    f = NamedFunction(uri,
                                      e.value,
                                      self.functions[fun_uri.EVAL],
                                      args=args)
                self.add_function(uri, f)



from aiddl_core.representation.int import Int


class AbsoluteValue:
    def __call__(self, x):
        if x.is_negative():
            return x * Int(-1)
        return x


class Addition:
    def __call__(self, args):
        s = Int(0)
        for i in range(0, args.size()):
            s += args[i]
        return s


class Subtraction:
    def __call__(self, args):
        if args.size() == 0:
            return Int(0)
        if args.size() == 1:
            return Int(0) - args[0]
        s = args[0]
        for i in range(1, args.size()):
            s -= args[i]
        return s


class Multiplication:
    def __call__(self, args):
        prod = Int(1)
        for i in range(0, args.size()):
            prod *= args[i]
        return prod


class Division:
    def __call__(self, args):
        if args.size() == 0:
            return Int(1)
        if args.size() == 1:
            return Int(1) / args[0]
        prod = args[0]
        for i in range(1, args.size()):
            prod /= args[i]
        return prod


class Modulo:
    def __call__(self, args):
        return Int(args.get(0).int_value % args.get(1).int_value)



from aiddl_core.representation.term import Term
from aiddl_core.representation.sym import Sym
from aiddl_core.representation.substitution import Substitution
from aiddl_core.representation.tuple import Tuple


class Entry:
    __slots__ = ["_type", "_name", "_value"]

    def __init__(self, entry_type, entry_name, entry_value):
        """ Create a new entry

        :param entry_type: type of the entry
        :param entry_name: name of the entry
        :param entry_value: value term of the entry
        """
        if entry_type is None or entry_name is None or entry_value is None:
            raise ValueError("Entry constructor needs non-null arguments.")
        self.__setattr__("_type", entry_type)
        self.__setattr__("_name", entry_name)
        self.__setattr__("_value", entry_value)

    @property
    def type(self) -> Term:
        """ The type of this entry

        :return: term representing the entry type
        """
        return self._type

    @property
    def name(self) -> Term:
        """ The name of this entry

        :return: term representing the entry name
        """
        return self._name

    @property
    def value(self) -> Term:
        """ The value of this entry

        :return: value term of this entry
        """
        return self._value

    def as_tuple(self) -> Tuple:
        """ Convert this entry to a tuple

        :return: tuple term with the same contents as the entry
        """
        return Tuple(self._type, self._name, self._value)

    def substitute(self, s: Substitution):
        """ Substitute the type, name, and value of this entry

        :param s: a substitution
        :return: new entry with the substitution applied
        """
        return Entry(self._type.substitute(s),
                     self._name.substitute(s),
                     self._value.substitute(s))

    def __hash__(self):
        return self._type.__hash__() \
            + 3*self._name.__hash__() \
            + 7*self._value.__hash__()

    def __eq__(self, other: object):
        if isinstance(other, Entry):
            return other.type == self._type \
                and other.name == self._name \
                and other.value == self._value
        return False

    def __ne__(self, other: object):
        return not self == other

    def __str__(self) -> str:
        return f"( {self.type} {self.name} {self.value})"

    def __repr__(self) -> str:
        return str(self._value)

    def unpack(self) -> object:
        """ Unpack entry and its contents in to a python tuple

        :return: python tuple with unpacked entry contents
        """
        return (self._type.unpack(),
                self._name.unpack(),
                self._value.unpack())



from aiddl_core.function.function import LazyFunctionMixin
from aiddl_core.representation import Term
from aiddl_core.representation.sym import Sym
from aiddl_core.representation.tuple import Tuple
from aiddl_core.representation.collection import Collection
from aiddl_core.representation.set import Set
from aiddl_core.representation.list import List
from aiddl_core.representation.entref import EntRef
from aiddl_core.representation.keyval import KeyVal

import aiddl_core.function as furi

SELF = Sym("#self")
SELF_ALT = Sym("#arg")


class Evaluator:
    def __init__(self, container):
        self._log_indent = ""
        self._verbose = False

        self._follow_references = False
        self._container = container

    def set_follow_references(self, flag):
        self._follow_references = flag

    def set_verbose(self, flag):
        self._verbose = flag

    def __call__(self, arg):
        if self._verbose:
            print(self._log_indent + str(arg))
            self._log_indent += "  "

        operator = None

        if not isinstance(arg, Tuple) or len(arg) == 0:
            if isinstance(arg, Collection):
                new_col = []
                for t in arg:
                    new_col.append(self(t))

                if isinstance(arg, List):
                    result = List(new_col)
                else:
                    result = Set(new_col)
            elif isinstance(arg, KeyVal):
                result = KeyVal(
                    self(arg.key), self(arg.value))
            elif isinstance(arg, EntRef):
                if self._follow_references:
                    result = self(self._container.resolve_reference(arg))
                else:
                    result = self._container.resolve_reference(arg)
            else:
                result = arg
        else:
            operator = arg.get(0)
            if isinstance(operator, EntRef):
                if isinstance(operator.target, Sym):
                    uri = operator.convert2uri()
                    if self._container._fun_reg.has_function(uri):
                        operator = uri
                    else:
                        operator = self._container.resolve_reference(operator)
                else:
                    operator = self._container.resolve_reference(operator)

            resolved_arguments = []
            processed_op = self(operator)
            is_lazy = False
            if not self._container._fun_reg.has_function(processed_op):
                resolved_arguments.append(processed_op)
            else:
                is_lazy = isinstance(self._container._fun_reg.get_function(operator), LazyFunctionMixin)
            for i in range(1, arg.size()):
                if not is_lazy:
                    if isinstance(arg.get(i), EntRef):
                        res_arg = self._container.resolve_reference(arg.get(i))
                    else:
                        res_arg = arg.get(i)

                    if not is_lazy:
                        resolved_arguments.append(self(res_arg))
                    else:
                        resolved_arguments.append(res_arg)
                else:
                    resolved_arguments.append(arg.get(i))
            if len(resolved_arguments) == 1:
                processed_args = resolved_arguments[0]
            else:
                processed_args = Tuple(resolved_arguments)
            if self._container._fun_reg.has_function(processed_op):
                result = self._container._fun_reg.get_function(processed_op)(processed_args)
            else:
                result = processed_args
 
        if self._verbose:
            if operator is None:
                operator = Sym("-")
            self._log_indent = self._log_indent[0:-2]
            print(self._log_indent + str(result) + "//" + str(operator))
        assert isinstance(result, Term)
        return result




from abc import abstractmethod

from .term import Term


class Num(Term):
    @abstractmethod
    def __add__(self, other: 'Num') -> 'Num':
        """ Add numerical term to another numerical term.

        :param other: term to add to this term
        :return: result of addition
        """

    @abstractmethod
    def __sub__(self, other: 'Num') -> 'Num':
        """ Subtract numerical term from this numerical term.

        :param other: term to subtract from this term
        :return: result of subtraction
        """

    @abstractmethod
    def __mul__(self, other: 'Num') -> 'Num':
        """ Multiply numerical term with this numerical term.

        :param other: term to multiply with this term
        :return: result of multiplication
        """

    @abstractmethod
    def __truediv__(self, other: 'Num') -> 'Num':
        """Divide this term by another numerical term.

        :param other: term to divide this term by
        :return: result of division
        """

    @abstractmethod
    def __floordiv__(self, other: 'Num') -> 'Num':
        """Divide this term by another numerical term and round to the nearest integer.

        :param other: term to divide this term by
        :return: result of division rounded down to the nearest integer
        """

    @abstractmethod
    def __lt__(self, other: 'Num') -> bool:
        """ Check if this term is less than another numerical term

        :param other: term to compare to
        :return: self < other
        """

    @abstractmethod
    def __le__(self, other: 'Num') -> bool:
        """ Check if this term is less than or equal to another numerical term

        :param other: term to compare to
        :return: self <= other
        """

    @abstractmethod
    def __gt__(self, other: 'Num') -> bool:
        """ Check if this term is greater than another numerical term

        :param other: term to compare to
        :return: self > other
        """

    @abstractmethod
    def __ge__(self, other: 'Num') -> bool:
        """ Check if this term is greater than or equal to another numerical term

        :param other: term to compare to
        :return: self >= other
        """

    @abstractmethod
    def is_zero(self) -> bool:
        """ Check if this term is zero

        :return: true if this numerical is equal to zero, false otherwise
        """

    @abstractmethod
    def is_positive(self) -> bool:
        """ Check if this term is positive

        :return: self > 0
        """

    @abstractmethod
    def is_negative(self) -> bool:
        """ Check if this term is positive

        :return: self < 0
        """

    def is_nan(self):
        """ Check if this term is not a number

        Note: this is required because NaN is not considered equal to NaN

        :return: true if this term is the non-a-number term, false otherwise
        """
        return False

    def is_inf(self):
        """ Check if this term is infinite

        Note: this is required because infinity is not considered equal to infinity

        :return: true if this term is positive or negative infinity, false otherwise
        """
        return False

    def is_inf_pos(self):
        """ Check if this term is positive infinity

        Note: this is required because infinity is not considered equal to infinity

        :return: true if this term is positive infinity, false otherwise
        """
        return False

    def is_inf_neg(self):
        """ Check if this term is positive infinity

        Note: this is required because infinity is not considered equal to infinity

        :return: true if this term is negative infinity, false otherwise
        """
        return False







class ComboIterator:
    def __init__(self, choices):
        self.slot_sizes = []
        self.slots = []
        self.carry = False
        self.done = False

        self.choices = choices

        for i in range(len(choices)):
            self.slot_sizes.append(len(self.choices[i]))
            self.slots.append(0)
        if len(self.choices) == 0:
            self.done = True
            return
        for choice in choices:
            if len(choice) == 0:
                self.done = True
                return

    def get_num_combos(self):
        n = 1
        for k in self.slot_sizes:
            n *= k
        return n

    def has_next(self):
        return not self.done

    def __next__(self):
        if self.done:
            raise StopIteration
        combo = []
        for i in range(0, len(self.slots)):
            combo.append(self.choices[i][self.slots[i]])
        self.iterate()
        return combo

    def iterate(self):
        if self.done:
            self.done = False
            for i in range(len(self.slots)):
                self.slots[i] = 0
        self.carry = False
        self.slots[0] += 1
        for i in range(len(self.slots)):
            if self.carry:
                self.slots[i] += 1
                self.carry = False
            if self.slots[i] == self.slot_sizes[i]:
                self.slots[i] = 0
                self.carry = True
        if self.carry:
            self.done = True

    def __iter__(self):
        return self


__c = ComboIterator([[1, 2],
                     [3, 4]])
assert(list(__c) == [[1, 3], [2, 3], [1, 4], [2, 4]])



import abc
import aiddl_core.representation.term as term


class Collection(term.Term):
    @abc.abstractmethod
    def __add__(self, other: term.Term) -> 'Collection':
        """Add term to collection.

        :param other: Term to add to this collection
        :return: Collection including the added term
        """

    @abc.abstractmethod
    def add_all(self, other: 'Collection') -> 'Collection':
        """Add another collection to this collection.

        :param other: Collection to add to this collection
        :return: Collection including the added terms
        """

    @abc.abstractmethod
    def put(self, key: term.Term, value: term.Term) -> 'Collection':
        """ Add key-value pair to collection.

        :param key: key term
        :param value: value term
        :return: this collection with a key value pair added
        """

    @abc.abstractmethod
    def put_all(self, other: 'Collection') -> 'Collection':
        """ Add collection of key values to collection.

        :param other: collection of key-value pairs
        :return: this collection with a key value pair added
        """

    @abc.abstractmethod
    def is_unique_map(self) -> bool:
        """ Check if this collection is a unique map (i.e., each key-value term has a different key)

        :return: true if each key-value pair in this collection has a different key, false otherwise
        """

    @abc.abstractmethod
    def is_empty(self) -> bool:
        """ Check if collection is empty.

        :return: true if this collection is empty, false otherwise
        """

    @abc.abstractmethod
    def remove(self, other: 'term.Term') -> 'Collection':
        """ Remove term from collection.

        :param other: term to remove
        :return: collection without the term
        """

    @abc.abstractmethod
    def remove_all(self, other: 'Collection') -> 'Collection':
        """ Remove collection of terms from collection.

        :param other: collection of terms to remove
        :return: collection without the terms
        """

    @abc.abstractmethod
    def __contains__(self, other: 'term.Term') -> 'bool':
        """ Check if a term is contained in this collection

        :param other: term to check
        :return: true if this collection contains the other term, false otherwise
        """

    def contains(self, other: 'term.Term') -> 'bool':
        """ Check if a term is contained in this collection

        :param other: term to check
        :return: true if this collection contains the other term, false otherwise
        """
        return other in self

    def contains_all(self, other: 'Collection') -> 'bool':
        """Check if collection contains another collection.

        :param other: collection
        :return: true if all terms in the other collection are also in this collection, false otherwise
        """
        for t in other:
            if t not in self:
                return False
        return True

    def contains_any(self, other: 'Collection') -> 'bool':
        """Check if collection contains at least one element of another collection.

        :param other: collection
        :return: true if any term in the other collection js also in this collection, false otherwise
        """
        for t in other:
            if t in self:
                return True
        return False

    @abc.abstractmethod
    def contains_key(self, other: 'term.Term') -> 'bool':
        """ Check if this collection term contains a key value with a specific key.

        :param other: a key term to search for
        :return: true if this collection contains a key-value pair with the specified key, false otherwise
        """



import os
import re
import sys
from pathlib import Path

from aiddl_core.representation.term import Term
from aiddl_core.representation.sym import Sym
from aiddl_core.representation.sym import Boolean
from aiddl_core.representation.str import Str
from aiddl_core.representation.var import Var
from aiddl_core.representation.rat import Rat
from aiddl_core.representation.real import Real
from aiddl_core.representation.int import Int
from aiddl_core.representation.inf import Inf
from aiddl_core.representation.tuple import Tuple
from aiddl_core.representation.set import Set
from aiddl_core.representation.list import List
from aiddl_core.representation.entref import EntRef
from aiddl_core.representation.keyval import KeyVal
from aiddl_core.representation.nan import NaN
from aiddl_core.representation.funref import FunRef
from aiddl_core.container.container import Entry


OTUPLE = "("
CTUPLE = ")"

OSET = "{"
CSET = "}"

OLIST = "["
CLIST = "]"

REF = "@"
SREF = "$"
FREF = "^"
ASSOC = ":"

MOD = Sym("#mod")
REQ = Sym("#req")
NAMES = Sym("#namespace")
NAMES_ALT = Sym("#nms")

QMARKS = '"'
ANY = "_"

SPECIAL = [OTUPLE, CTUPLE,
           OSET, CSET,
           OLIST, CLIST,
           ASSOC, REF, SREF, FREF]

INFINITY = ["INF", "+INF", "-INF"]
NAN = "NaN"

REQ = Sym("#req")
MOD = Sym("#mod")


def is_float(s):
    try:
        float(s)
        return "." in s
    except ValueError:
        return False


def is_bin(s):
    if s[0:2] == "#b":
        try:
            tmp = s[2:len(s)]
            int(tmp, 2)
            return True
        except ValueError:
            return False
    return False


def is_hex(s):
    if s[0:2] == "#x":
        try:
            tmp = s[2:len(s)]
            int(tmp, 16)
            return True
        except ValueError:
            return False
    return False


def is_oct(s):
    if s[0:2] == "#o":
        try:
            tmp = s[2:len(s)]
            int(tmp, 8)
            return True
        except ValueError:
            return False
    return False


def is_int(s):
    try:
        int(s)
        return "." not in s and "e" not in s
    except ValueError:
        return False


def pop(s):
    r = s[-1]
    del s[-1]
    return r


def peek(s):
    return s[-1]


def push(s, a):
    s.append(a)


def parse_string(s, aiddl_paths, freg, my_folder='./'):
    str_id = 0
    str_lookup = {}

    bs = False
    current_str_start = None
    slices = []
    s_new = ""
    for i in range(len(s)):
        c = s[i]
        if c == '"' and not bs and current_str_start is not None:
            str_lookup[str_id] = s[current_str_start:i]

            slices.append((current_str_start, i, str(str_id)))
            s_new += ' "%d"' % (str_id)
            current_str_start = None
            str_id += 1
        elif c == '"' and not bs:
            current_str_start = i+1
        elif current_str_start is None:
            s_new += c

        if c == '\\':
            bs = True
        else:
            bs = False

    # for sl in slices:
    #     s = s[:sl[0]] + sl[2] + s[sl[1]:]

    s = s_new

    s = re.sub(r";;.*\n", "", s)
    s = s.replace(",", " ")
    s = s.replace("\n", " ")
    s = s.replace("\t", " ")

    for token in SPECIAL:
        s = s.replace(token, " " + token + " ")
    while "  " in s:
        s = s.replace("  ", " ")
    s = s.strip()
    stack = []
    module_name = None
    self_ref = None
    local_refs = {}
    tokens = s.split(" ")
    tuple_depth = 0
    for i in range(len(tokens)):
        token = tokens[i].strip()
        if token == CLIST:
            assembled_list = []
            current = pop(stack)
            while current != OLIST:
                assembled_list.append(current)
                current = pop(stack)
            assembled_list.reverse()
            term = List(assembled_list)
            # push(stack, term)
        elif token == CTUPLE:
            tuple_depth -= 1
            assembled_list = []
            current = pop(stack)
            while current != OTUPLE:
                assembled_list.append(current)
                current = pop(stack)
            assembled_list.reverse()
            term = Tuple(assembled_list)
            # push(stack, term)
            if tuple_depth == 0:
                if term.get(0) == MOD:
                    self_ref = term.get(1)
                    module_name = term.get(2)
                elif term.get(0) == REQ or \
                     term.get(0) == NAMES or \
                     term.get(0) == NAMES_ALT:
                    local_refs[term.get(1)] = term.get(2)
                    if isinstance(term.get(2), Str):
                        fname = term.get(2).string
                        req_mod_n = get_mod_name_from_file(my_folder + fname, aiddl_paths)
                        if req_mod_n is not None:
                            local_refs[term.get(1)] = req_mod_n
        elif token == CSET:
            assembled_set = []
            current = pop(stack)
            while current != OSET:
                assembled_set.append(current)
                current = pop(stack)
            term = Set(assembled_set)
        elif token in SPECIAL:
            push(stack, token)
            if token == OTUPLE:
                tuple_depth += 1
            continue
        else:
            term = None
            if token[0] == "?":
                term = Var(name=token)
            elif token == "_":
                term = Var()
            elif token[0] == '"':
                string = str_lookup[int(token[1:-1])]
                term = Str(string)
            elif "/" in token:
                n = token.split("/")[0]
                d = token.split("/")[1]
                term = Rat(int(n), int(d))
            elif is_float(token):
                term = Real(float(token))
            elif is_int(token):
                term = Int(int(token))
            elif is_bin(token):
                term = Int(int(token[2:], 2))
            elif is_oct(token):
                term = Int(int(token[2:], 8))
            elif is_hex(token):
                term = Int(int(token[2:], 16))
            elif token in INFINITY:
                if "-" in token:
                    term = Inf.neg()
                else:
                    term = Inf.pos()
            elif token == NAN:
                term = NaN()
            elif token == "true":
                term = Boolean(True)
            elif token == "false":
                term = Boolean(False)
            else:
                term = Sym(token)
        if len(stack) > 0 and i+1 < len(tokens) and tokens[i+1] != "@":
            back_resolve = True
            while back_resolve and len(stack) > 0:
                back_resolve = False
                if peek(stack) == SREF:
                    pop(stack)
                    if not isinstance(term, KeyVal):
                        term = EntRef(term, module_name)
                    else:
                        val = term.get_value()
                        key = EntRef(term.get_key(), module_name)
                        term = KeyVal(key, val)
                    back_resolve = True
                elif peek(stack) == REF:
                    pop(stack)
                    name = pop(stack)
                    if not isinstance(name, FunRef):
                        if term == self_ref:
                            # print("Pointing to %s in module %s (aka %s)"
                            #       % (str(name), str(module_name), str(term)))
                            term = EntRef(name, module_name, alias=term)
                        else:
                            # print("Pointing to %s in module %s (aka %s)" % (str(name), str(module_name), str(term)))
                            term = EntRef(name,
                                          local_refs[term],
                                          alias=term)
                    else:
                        bad_fref = name.get_fref()
                        if term == self_ref:
                            term = FunRef(bad_fref, freg,
                                          module=module_name)
                        else:
                            term = FunRef(bad_fref, freg,
                                          module=module_name)
                    back_resolve = True
                elif peek(stack) == FREF:
                    pop(stack)
                    term = FunRef(term, freg)
                    back_resolve = True
                elif peek(stack) == ASSOC:
                    pop(stack)
                    key = pop(stack)
                    if isinstance(key, KeyVal):
                        term = KeyVal(key._key,
                                      KeyVal(key._value, term))
                    else:
                        term = KeyVal(key, term)
                    back_resolve = True
        push(stack, term)
    return (stack, module_name, self_ref, local_refs)


def parse_term(s):
    return parse_string(s, None, None)[0][0]


def to_aiddl(o):
    if isinstance(o, Term):
        return o
    if isinstance(o, str):
        return Str(o)
    elif isinstance(o, int):
        return Int(o)
    elif isinstance(o, float):
        return Real(o)
    elif isinstance(o, list):
        List([to_aiddl(x) for x in o])
    elif isinstance(o, set):
        Set([to_aiddl(x) for x in o])
    else:
        raise ValueError("Not convertable to term:", type(o), o)


def find_and_open_file(filename, aiddl_paths):
    if os.path.isfile(filename):
        f = open(filename, "r")
        return f
    else:
        for path in aiddl_paths:
            fname = str(path) + "/" + str(filename)
            if os.path.isfile(fname):
                f = open(fname, "r")
                return f
        raise ValueError(
            "File: " + str(filename) +
            " not found in current directory or AIDDL_PATH: "
            + str(aiddl_paths))


def get_mod_name_from_file(fname, aiddl_paths):
    f = find_and_open_file(fname, aiddl_paths)
    s = ""
    in_mod_str = False
    found_mod_entry = False
    for line in f.readlines():
        line = re.sub(r";;.*\n", "", line)
        if "#mod" in line:
            in_mod_str = True
        if in_mod_str:
            s += line
            if ")" in line:
                found_mod_entry = True
                f.close()
                break

    if found_mod_entry:
        req_mod_n = s.split("(")[1].split(")")[0].split()[-1]
        return Sym(req_mod_n)
    else:
        return None


def is_known_module(mod_name):
    lookup = get_mod_file_lookup(collect_aiddl_paths([]))
    return mod_name in lookup.keys()


def get_mod_file_lookup(paths):
    m = {}
    # if "win32" == sys.platform:
    #     paths += os.environ['AIDDL_PATH'].split(";")
    # else:
    #     paths += os.environ['AIDDL_PATH'].split(":")
    for path in paths:
        for f_path in Path(path).rglob('*.aiddl'):
            if "#" not in str(f_path):
                f_name = f_path.resolve()
                mod_name = get_mod_name_from_file(f_name, paths)
                if mod_name is not None:
                    m[mod_name] = f_name
    return m


def collect_aiddl_paths(paths):
    paths = list(paths)
    if "win32" == sys.platform:
        paths += os.environ['AIDDL_PATH'].split(";")
    else:
        paths += os.environ['AIDDL_PATH'].split(":")
    return paths


# def parse(filename, container, freg, current_folder, root_folders=[]):
#     mod = parse_internal(filename, container, freg, current_folder, root_folders=root_folders)
#     container.toggle_namespaces(True)
#     freg.load_def(container)
#     freg.load_type_functions(container)
#     freg.load_container_interfaces(container)
#     # freg.load_req_python_functions(container)
#     return mod


def parse(filename, container, root_folders=[]):
    abs_file_path = os.path.abspath(filename)
    mod = parse_internal(abs_file_path, container, root_folders=root_folders)
    container.toggle_namespaces(True)
    container._fun_reg.load_def(container)
    container._fun_reg.load_type_functions(container)
    container._fun_reg.load_container_interfaces(container)
    # freg.load_req_python_functions(container)
    return mod


def parse_internal(filename, container, root_folders=[]):
    current_folder = os.path.dirname(os.path.abspath(filename)) + "/"
    # print("Parsing:", filename)
    aiddl_paths = collect_aiddl_paths(root_folders)
    mod_name_lookup = get_mod_file_lookup(aiddl_paths)
    f = find_and_open_file(filename, aiddl_paths)
    # f_current_folder = os.path.dirname(filename)

    s = f.read()
    f.close()
    terms, mod_name, self_ref, local_refs = parse_string(s,
                                                         aiddl_paths,
                                                         container._fun_reg,
                                                         my_folder=current_folder)

    mod_entry = terms[0]
    container.add_module(mod_name)
    container.add_module_alias(mod_name, mod_entry.get(1), mod_name)
    for term in terms[0:]:
        assert isinstance(term, Tuple) and len(term) == 3
        if term.get(0) == REQ or term.get(0) == NAMES or term.get(0) == NAMES_ALT:
            # print("Loading requirement or namespace: ", term)
            # print(os.environ['AIDDL_PATH'])
            fname = None
            if isinstance(term.get(2), Str):
                fname = current_folder + term.get(2).string
            elif isinstance(term.get(2), Sym):
                # print(mod_name_lookup)
                # for s in mod_name_lookup.keys():
                #     print(s, "->", mod_name_lookup[s])
                fname = mod_name_lookup[term.get(2)]

            if fname is not None:
                req_mod_name = parse_internal(fname,
                                              container,
                                              root_folders=root_folders)
                container.add_module_alias(mod_name, term.get(1), req_mod_name)
                entry = Entry(term.get(0), term.get(1), req_mod_name)
                container.set_entry(entry, module=mod_name)
            else:
                entry = Entry(term.get(0), term.get(1), term.get(2))
                container.set_entry(entry, module=mod_name)
        else:
            entry = Entry(term.get(0), term.get(1), term.get(2))
            container.set_entry(entry, module=mod_name)
    return mod_name



import aiddl_core.representation.term as term
import aiddl_core.representation.num as numerical
import aiddl_core.representation.int as integer
import aiddl_core.representation.real as real
import aiddl_core.representation.inf as infinity
import aiddl_core.representation.nan as nan


from math import floor


class Rat(numerical.Num):
    __slots__ = ["_p", "_q"]

    def __init__(self, n, d):
        if d == 0:
            raise AttributeError("Rational number with zero denominator.")
        gcd = Rat._gcd(n, d)
        super(term.Term, self).__setattr__("_p", n//gcd)
        super(term.Term, self).__setattr__("_q", d//gcd)

    @property
    def nominator(self):
        return self._p

    @property
    def denominator(self):
        return self._q

    @property
    def int_value(self):
        return self._p // self._q

    @property
    def real_value(self):
        return self._p / self._q

    @staticmethod
    def _gcd(a, b):
        if b == 0:
            return a
        else:
            return Rat._gcd(b, a % b)

    def __add__(self, other):
        if isinstance(other, Rat):
            return Rat(self._p * other._q + other._p * self._q,
                       self._q * other._q)
        elif isinstance(other, integer.Int):
            return Rat(self._p + self._q * other._value, self._q)
        elif isinstance(other, real.Real):
            return real.Real(self._p / self._q + other._value)
        elif isinstance(other, infinity.Inf):
            return other
        elif isinstance(other, nan.NaN):
            return other
        raise AttributeError("Not a numerical term: " + str(other))

    def __sub__(self, other):
        if isinstance(other, Rat):
            return Rat(self._p * other._q - other._p * self._q,
                       self._q * other._q)
        elif isinstance(other, integer.Int):
            return Rat(self._p - self._q * other._value, self._q)
        elif isinstance(other, real.Real):
            return real.Real(self._p / self._q - other._value)
        elif isinstance(other, infinity.Inf):
            return infinity.Inf(not other._is_positive)
        elif isinstance(other, nan.NaN):
            return other
        raise AttributeError("Not a numerical term: " + str(other))

    def __mul__(self, other):
        if isinstance(other, Rat):
            return Rat(self._p * other._p, self._q * other._q)
        elif isinstance(other, integer.Int):
            return Rat(self._p * other._value, self._q)
        elif isinstance(other, real.Real):
            return real.Real(self._p / self._q * other._value)
        elif isinstance(other, infinity.Inf):
            if self == integer.Int(0):
                return nan.NaN()
            else:
                return infinity.Inf(other._is_positive ==
                                    (self > integer.Int(0)))
        elif isinstance(other, nan.NaN):
            return other
        raise AttributeError("Not a numerical term: " + str(other))

    def __truediv__(self, other):
        if isinstance(other, Rat):
            return Rat(self._p * other._q, self._q * other._p)
        elif isinstance(other, integer.Int):
            return Rat(self._p, self._q * other._value)
        elif isinstance(other, real.Real):
            return real.Real((self._p / self._q) / other._value)
        elif isinstance(other, infinity.Inf):
            return integer.Int(0)
        elif isinstance(other, nan.NaN):
            return other
        raise AttributeError("Not a numerical term: " + str(other))

    def __floordiv__(self, other):
        if isinstance(other, Rat):
            return integer.Int((self._p * other._q)
                               // (self._q * other._p))
        elif isinstance(other, integer.Int):
            return integer.Int(self._p // (self._q * other._value))
        elif isinstance(other, real.Real):
            return integer.Int(int(floor(self._p / self._q)
                                   / other._value))
        elif isinstance(other, infinity.Inf):
            return integer.Int(0)
        elif isinstance(other, nan.NaN):
            return other
        raise AttributeError("Not a numerical term: " + str(other))

    def __str__(self):
        return str(self._p)+"/"+str(self._q)

    def __eq__(self, other):
        if isinstance(other, Rat):
            return self._p*other._q == other._p*self._q
        elif isinstance(other, integer.Int):
            return self._p == other._value*self._q
        elif isinstance(other, real.Real):
            return (self._p / self._q) == other._value
        elif isinstance(other, infinity.Inf):
            return False
        else:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __lt__(self, other):
        if isinstance(other, Rat):
            return self._p*other._q < other._p*self._q
        elif isinstance(other, integer.Int):
            return self._p < other._value*self._q
        elif isinstance(other, real.Real):
            return (self._p / self._q) < other._value
        elif isinstance(other, infinity.Inf):
            return other._is_positive
        elif isinstance(other, nan.Nan):
            return False
        raise AttributeError("Not a numerical term: " + str(other))

    def __le__(self, other):
        if isinstance(other, Rat):
            return self._p*other._q <= other._p*self._q
        elif isinstance(other, integer.Int):
            return self._p <= other._value*self._q
        elif isinstance(other, real.Real):
            return (self._p / self._q) <= other._value
        elif isinstance(other, infinity.Inf):
            return other._is_positive
        elif isinstance(other, nan.Nan):
            return False
        raise AttributeError("Not a numerical term: " + str(other))

    def __gt__(self, other):
        if isinstance(other, Rat):
            return self._p*other._q > other._p*self._q
        elif isinstance(other, integer.Int):
            return self._p > other._value*self._q
        elif isinstance(other, real.Real):
            return (self._p / self._q) > other._value
        elif isinstance(other, infinity.Inf):
            return not other._is_positive
        elif isinstance(other, nan.Nan):
            return False
        raise AttributeError("Not a numerical term: " + str(other))

    def __ge__(self, other):
        if isinstance(other, Rat):
            return self._p*other._q >= other._p*self._q
        elif isinstance(other, integer.Int):
            return self._p >= other._value*self._q
        elif isinstance(other, real.Real):
            return (self._p / self._q) >= other._value
        elif isinstance(other, infinity.Inf):
            return not other._is_positive
        elif isinstance(other, nan.Nan):
            return False
        raise AttributeError("Not a numerical term: " + str(other))

    def is_zero(self):
        return self._p == 0

    def is_positive(self):
        return self._p > 0

    def is_negative(self):
        return self._p < 0

    def __hash__(self):
        return hash(self._p) + hash(self._q)

    def unpack(self):
        return str(self._p) + "/" + str(self._q)



import aiddl_core.representation.term as term
from aiddl_core.representation.collection import Collection
from aiddl_core.representation.keyval import KeyVal


class Set(Collection):
    __slots__ = ["_internal_set", "_internal_map"]

    def __init__(self, *args):
        internal_set = set()
        internal_map = {}
        for t in args:
            if not isinstance(t, term.Term):
                for e in t:
                    internal_set.add(e)
                    if isinstance(e, KeyVal):
                        internal_map[e.key] = e.value
            else:
                internal_set.add(t)
                if isinstance(t, KeyVal):
                    internal_map[t.key] = t.value
        super(term.Term, self).__setattr__("_internal_set", internal_set)
        super(term.Term, self).__setattr__("_internal_map", internal_map)
        super(term.Term, self).__setattr__("_hash", hash(tuple(internal_set)))

    def resolve(self, container):
        l_new = []
        change = False
        for e in self._internal_set:
            e_new = e.resolve(container)
            l_new.append(e_new)
            change = change or (e != e_new)
        if not change:
            return self
        else:
            return Set(l_new)

    def substitute(self, substitution):
        l_new = []
        change = False
        for e in self._internal_set:
            e_new = e.substitute(substitution)
            l_new.append(e_new)
            change = change or (e != e_new)
        if not change:
            return self
        else:
            return Set(l_new)

    def match(self, other):
        raise ValueError("Set term cannot be on left side of match.")

    def __iter__(self):
        return self._internal_set.__iter__()

    def __next__(self):
        pass

    def add(self, value):
        l_new = set()
        for t in self._internal_set:
            l_new.add(t)
        l_new.add(value)
        return Set(l_new)

    def __add__(self, S):
        return self.add(S)
        
    def add_all(self, S):
        l_new = set()
        for t in self._internal_set:
            l_new.add(t)
        for e in S:
            l_new.add(e)
        return Set(l_new)

    def put(self, key, value):
        l_new = set()
        for t in self._internal_set:
            if not isinstance(t, KeyVal) or t.key != key:
                l_new.add(t)
        l_new.add(KeyVal(key, value))
        return Set(l_new)

    def put_all(self, values):
        l_new = set()
        for k in self._internal_map.keys():
            if not values.contains_key(k):
                l_new.add(KeyVal(k, self.get(k)))
        for t in values:
            if isinstance(t, KeyVal):
                l_new.add(t)
        return Set(l_new)

    def __len__(self):
        return len(self._internal_set)

    def is_empty(self):
        return len(self._internal_set) == 0

    def remove(self, e):
        l_new = []
        for t in self._internal_set:
            if not e == t:
                l_new.append(t)
        return Set(l_new)

    def remove_all(self, c):
        l_new = []
        for t in c:
            if t not in c:
                l_new.append(t)
        return Set(l_new)

    def get(self, n):
        if n in self._internal_map.keys():
            return self._internal_map[n]
        return None

    def is_unique_map(self):
        seen = set()
        for e in self._internal_set:
            if not isinstance(e, KeyVal) or e.key in seen:
                return False
            seen.add(e.key)
        return True

    def __contains__(self, other):
        return other in self._internal_set

    def contains_key(self, other):
        return other in self._internal_map.keys()

    def __eq__(self, other):
        return isinstance(other, Set) \
            and self._internal_set == other._internal_set

    def __ne__(self, other):
        return not self == other

    def __str__(self):
        if len(self._internal_set) == 0:
            return "{}"
        s = "{"
        for e in self._internal_set:
            s += str(e) + " "
        return s[0:-1] + "}"

    def __repr__(self):
        return self.__str__()

    def __hash__(self):
        return self._hash

    def unpack(self):
        s = set()
        for e in self:
            s.add(e.unpack())
        return s
